from requre.objects import ObjectStorage
from requre.storage import PersistentObjectStorage


@ObjectStorage.decorator_plain
def return_result():
    return {"value": input("insert value: ")}


PersistentObjectStorage().storage_file = "values.yaml"
value = return_result()

ObjectStorage.persistent_storage.dump()
print(PersistentObjectStorage().storage_object)
print(value)



# MIT License
#
# Copyright (c) 2018-2019 Red Hat, Inc.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import os
from typing import Any, Callable

from git.refs.head import HEAD
from git.remote import PushInfo
from git.repo.base import Repo
from git.util import IterableList

from requre.objects import ObjectStorage
from requre.helpers.files import StoreFiles


class PushInfoStorageList(ObjectStorage):
    __ignored = ["_remote"]
    __response_keys_special = ["local_ref"]
    __response_keys = list(
        set(PushInfo.__slots__) - set(__ignored) - set(__response_keys_special)
    )

    object_type = IterableList

    def to_serializable(self, obj: Any) -> Any:
        output = list()
        for item in obj:
            tmp = dict()
            for key in self.__response_keys:
                tmp[key] = getattr(item, key)
            for key in self.__response_keys_special:
                if key == "local_ref":
                    tmp[key] = [
                        getattr(item, key).repo.git_dir,
                        getattr(item, key).path,
                    ]
            output.append(tmp)
        return output

    def from_serializable(self, data: Any) -> Any:
        out = self.object_type("name")
        for item in data:
            tmp = PushInfo(None, None, None, None)
            for key in self.__response_keys:
                setattr(tmp, key, item[key])
            for key in self.__response_keys_special:
                if key == "local_ref":
                    setattr(tmp, key, HEAD(Repo(item[key][0])))
            out.append(tmp)
        return out

    @classmethod
    def execute(cls, keys: list, func: Callable, *args, **kwargs) -> Any:
        """
        Class method to store or read object from persistent storage
        with explicit set of *args, **kwargs parameters to use as keys
        :param keys: list of keys used as parameter
        :param func: original function
        :param args: parameters of original function
        :param kwargs: parameters of original function
        :return: output of called func
        """

        output = super().execute(keys, func, *args, **kwargs)
        git_object = args[0]
        remote_url = git_object.repo.remotes[git_object.name].url
        if os.path.isdir(remote_url):
            StoreFiles.explicit_reference(remote_url)
        return output



import importlib

from requre.helpers.requests_response import RequestResponseHandling
from requre.storage import PersistentObjectStorage
from requre.utils import StorageMode
from tests.testbase import BaseClass


class StoreAnyRequest(BaseClass):
    domain = "https://example.com/"

    def setUp(self) -> None:
        super().setUp()
        self.requests = importlib.import_module("requests")
        self.post_orig = getattr(self.requests, "post")

    def tearDown(self) -> None:
        super().tearDown()
        setattr(self.requests, "post", self.post_orig)

    def testRawCall(self):
        """
        Test if is class is able to explicitly write and read request handling
        """
        keys = [self.domain]
        sess = RequestResponseHandling(store_keys=keys)
        response = self.requests.post(*keys)
        sess.write(response)

        response_after = sess.read()
        self.assertIsInstance(response_after, self.requests.models.Response)
        self.assertNotIn("Example Domain", str(sess.persistent_storage.storage_object))
        self.assertIn("Example Domain", response_after.text)

    def testExecuteWrapper(self):
        """
        test if it is able to use explicit decorator_all_keys for storing request handling
        :return:
        """
        response_before = RequestResponseHandling.execute_all_keys(
            self.requests.post, self.domain
        )
        PersistentObjectStorage().dump()
        PersistentObjectStorage().mode = StorageMode.read
        response_after = RequestResponseHandling.execute_all_keys(
            self.requests.post, self.domain
        )
        self.assertEqual(response_before.text, response_after.text)
        self.assertRaises(
            Exception,
            RequestResponseHandling.execute_all_keys,
            self.requests.post,
            self.domain,
        )

    def testFunctionDecorator(self):
        """
        Test main purpose of the class, decorate post function and use it then
        """
        self.requests.post = RequestResponseHandling.decorator_all_keys(
            self.requests.post
        )
        response_before = self.requests.post(self.domain)
        PersistentObjectStorage().dump()
        PersistentObjectStorage().mode = StorageMode.read

        response_after = self.requests.post(self.domain)
        self.assertEqual(response_before.text, response_after.text)
        self.assertRaises(Exception, self.requests.post, self.domain)

    def testFunctionDecoratorNotFound(self):
        """
        Check if it fails with Exception in case request is not stored
        """
        self.requests.post = RequestResponseHandling.decorator_all_keys(
            self.requests.post
        )
        PersistentObjectStorage().dump()
        PersistentObjectStorage().mode = StorageMode.read
        self.assertRaises(Exception, self.requests.post, self.domain, data={"a": "b"})

    def testFunctionCustomFields(self):
        """
        Test if it is able to use partial storing of args, kwargs
        prepare to avoid leak authentication to data
        """
        self.requests.post = RequestResponseHandling.decorator(item_list=[0])(
            self.requests.post
        )
        response_before = self.requests.post(self.domain)
        response_google_before = self.requests.post(
            "http://www.google.com", data={"a": "b"}
        )
        PersistentObjectStorage().dump()
        PersistentObjectStorage().mode = StorageMode.read

        response_after = self.requests.post(self.domain)
        response_google_after = self.requests.post("http://www.google.com")
        self.assertEqual(response_before.text, response_after.text)
        self.assertEqual(response_google_before.text, response_google_after.text)
        self.assertRaises(Exception, self.requests.post, self.domain)

    def testFunctionCustomFieldsWrong(self):
        """
        Check exceptions if using partial keys storing
        """
        self.requests.post = RequestResponseHandling.decorator(item_list=[0, "data"])(
            self.requests.post
        )
        self.requests.post(self.domain, data={"a": "b"})
        response_2 = self.requests.post(self.domain, data={"c": "d"})
        PersistentObjectStorage().dump()
        PersistentObjectStorage().mode = StorageMode.read

        self.assertRaises(Exception, self.requests.post, self.domain, data={"x": "y"})
        self.assertRaises(KeyError, self.requests.post, self.domain)
        response_2_after = self.requests.post(self.domain, data={"c": "d"})
        self.assertEqual(response_2.text, response_2_after.text)



from requre.objects import ObjectStorage
from requre.storage import PersistentObjectStorage, DataMiner
from requre.utils import StorageMode
from tests.testbase import BaseClass


class OwnClass:
    def __init__(self, num):
        self.num = num

    def get_num(self, return_num, extended=False):
        if extended:
            return f"{return_num}"
        else:
            return return_num


class StoreAnyRequest(BaseClass):
    domain = "https://example.com/"

    def testRawCall(self):
        """
        Test if is class is able to explicitly write and read object handling
        """
        keys = [1]
        sess = ObjectStorage(store_keys=keys)
        obj_before = OwnClass(*keys)
        sess.write(obj_before)
        obj_after = sess.read()
        self.assertIsInstance(obj_before, OwnClass)
        # it depickle object but object has not be same
        self.assertNotEqual(obj_before, obj_after)
        self.assertEqual(obj_before.num, obj_after.num)
        self.assertEqual(obj_after.num, 1)

    def testExecuteWrapper(self):
        """
        test if it is able to use explicit decorator_all_keys for storing object handling
        :return:
        """
        obj_before = ObjectStorage.execute_all_keys(OwnClass, 1)
        PersistentObjectStorage().dump()
        PersistentObjectStorage().mode = StorageMode.read
        obj_after = ObjectStorage.execute_all_keys(OwnClass, 1)
        self.assertEqual(obj_before.num, obj_after.num)
        # all objects are already read, next have to fail
        self.assertRaises(Exception, ObjectStorage.execute_all_keys, OwnClass, 1)

    def testFunctionDecorator(self):
        """
        Test main purpose of the class, decorate class and use it then
        """
        decorated_own = ObjectStorage.decorator_all_keys(OwnClass)
        obj_before = decorated_own(1)
        PersistentObjectStorage().dump()
        PersistentObjectStorage().mode = StorageMode.read
        obj_after = decorated_own(1)
        self.assertEqual(obj_before.num, obj_after.num)
        # all objects are already read, next have to fail
        self.assertRaises(Exception, decorated_own, 1)


class CallDebug(BaseClass):
    def setUp(self) -> None:
        super().setUp()
        DataMiner().store_arg_debug_metadata = True

    def tearDown(self) -> None:
        DataMiner().store_arg_debug_metadata = False
        super().tearDown()

    def testCallDebug(self):
        OwnClass.get_num = ObjectStorage.decorator_plain(OwnClass.get_num)
        test_obj = OwnClass(3)
        obj_before_4 = test_obj.get_num(4)
        obj_before_5 = test_obj.get_num(5, extended=True)
        PersistentObjectStorage().dump()
        PersistentObjectStorage().mode = StorageMode.read
        obj_after_4 = test_obj.get_num(4)
        obj_after_4_meta = DataMiner().metadata
        obj_after_5 = test_obj.get_num(5, extended=True)
        obj_after_5_meta = DataMiner().metadata
        self.assertEqual(obj_before_4, obj_after_4)
        self.assertIsInstance(obj_after_4, int)
        self.assertRegex(
            obj_after_4_meta[DataMiner().METADATA_ARG_DEBUG_KEY], "get_num(.*, 4)"
        )
        self.assertEqual(obj_before_5, obj_after_5)
        self.assertIsInstance(obj_after_5, str)
        self.assertRegex(
            obj_after_5_meta[DataMiner().METADATA_ARG_DEBUG_KEY],
            "get_num(.*, 5, extended=True)",
        )

    def testCallNoDebug(self):
        DataMiner().store_arg_debug_metadata = False
        OwnClass.get_num = ObjectStorage.decorator_plain(OwnClass.get_num)
        test_obj = OwnClass(3)
        obj_before_4 = test_obj.get_num(4)
        PersistentObjectStorage().dump()
        PersistentObjectStorage().mode = StorageMode.read
        obj_after_4 = test_obj.get_num(4)
        obj_after_4_meta = DataMiner().metadata
        self.assertEqual(obj_before_4, obj_after_4)
        self.assertIsInstance(obj_after_4, int)
        self.assertEqual(obj_after_4_meta.get(DataMiner().METADATA_ARG_DEBUG_KEY), None)



import os
import time

from requre.constants import VERSION_REQURE_FILE
from requre.exceptions import PersistentStorageException
from requre.helpers.simple_object import Simple
from requre.storage import (
    DataMiner,
    DataTypes,
    StorageKeysInspectDefault,
    StorageKeysInspectSimple,
    StorageKeysInspect,
    PersistentObjectStorage,
    StorageKeysInspectUnique,
)
from requre.utils import StorageMode
from tests.testbase import BaseClass


class Base(BaseClass):
    keys = ["a", "b"]

    def setUp(self) -> None:
        super().setUp()
        PersistentObjectStorage().dump_after_store = False

    def store_keys_example(self):
        PersistentObjectStorage().store(self.keys, values="c", metadata={})
        PersistentObjectStorage().store(self.keys, values="d", metadata={})

    def test_simple(self):
        self.store_keys_example()
        self.assertEqual("c", PersistentObjectStorage()[self.keys])
        self.assertEqual("d", PersistentObjectStorage()[self.keys])

    def test_exception(self):
        self.test_simple()
        self.assertRaises(
            PersistentStorageException, PersistentObjectStorage().read, self.keys
        )

    def test_write(self):
        self.store_keys_example()
        self.assertFalse(os.path.exists(self.response_file))
        PersistentObjectStorage().dump()
        self.assertTrue(os.path.exists(self.response_file))


class Versioning(BaseClass):
    def setUp(self) -> None:
        super().setUp()
        PersistentObjectStorage().dump_after_store = False

    def test_no_version(self):
        """
        Check if storage is able to read and write version to persistent storage file
        """
        self.assertEqual({}, PersistentObjectStorage().metadata)
        self.assertEqual(0, PersistentObjectStorage().storage_file_version)
        self.assertEqual(
            None,
            PersistentObjectStorage().metadata.get(
                PersistentObjectStorage().version_key
            ),
        )
        self.assertFalse(os.path.exists(self.response_file))

    def test_no_version_after_dump(self):
        PersistentObjectStorage().dump()
        PersistentObjectStorage().storage_object = {}
        self.assertEqual({}, PersistentObjectStorage().metadata)
        self.assertEqual(0, PersistentObjectStorage().storage_file_version)

    def test_current_version(self):
        PersistentObjectStorage().dump()
        self.assertTrue(os.path.exists(self.response_file))
        self.assertEqual(
            VERSION_REQURE_FILE, PersistentObjectStorage().storage_file_version
        )
        self.assertNotEqual({}, PersistentObjectStorage().metadata)

    def test_current_version_after_load(self):
        PersistentObjectStorage().dump()
        PersistentObjectStorage().load()
        self.assertEqual(
            VERSION_REQURE_FILE, PersistentObjectStorage().storage_file_version
        )
        self.assertNotEqual({}, PersistentObjectStorage().metadata)
        self.assertIn("version_storage_file", PersistentObjectStorage().metadata)


class TestStoreTypes(BaseClass):
    keys = ["a", "b"]

    def tearDown(self):
        DataMiner().data_type = DataTypes.List

    def test_default(self):
        self.assertEqual(DataTypes.List, DataMiner().data_type)

    def test_list_data(self):
        PersistentObjectStorage().store(keys=self.keys, values="x", metadata={})
        PersistentObjectStorage().store(keys=self.keys, values="y", metadata={})
        self.assertEqual(2, len(PersistentObjectStorage().storage_object["a"]["b"]))
        self.assertEqual("x", PersistentObjectStorage()[self.keys])
        self.assertEqual(1, len(PersistentObjectStorage().storage_object["a"]["b"]))
        self.assertEqual("y", PersistentObjectStorage()[self.keys])
        self.assertEqual(0, len(PersistentObjectStorage().storage_object["a"]["b"]))

    def test_value_data(self):
        DataMiner().data_type = DataTypes.Value
        PersistentObjectStorage().store(keys=self.keys, values="x", metadata={})
        PersistentObjectStorage().store(keys=self.keys, values="y", metadata={})
        self.assertEqual(2, len(PersistentObjectStorage().storage_object["a"]["b"]))
        self.assertEqual("y", PersistentObjectStorage()[self.keys])
        self.assertEqual(2, len(PersistentObjectStorage().storage_object["a"]["b"]))
        self.assertEqual("y", PersistentObjectStorage()[self.keys])
        self.assertEqual(2, len(PersistentObjectStorage().storage_object["a"]["b"]))

    def test_dict_data(self):
        DataMiner().data_type = DataTypes.Dict
        DataMiner().key = "first-key"
        PersistentObjectStorage().store(keys=self.keys, values="x", metadata={})
        DataMiner().key = "second-key"
        PersistentObjectStorage().store(keys=self.keys, values="y", metadata={})

        self.assertIn("first-key", PersistentObjectStorage().storage_object["a"]["b"])
        self.assertIn("second-key", PersistentObjectStorage().storage_object["a"]["b"])

        self.assertEqual("y", PersistentObjectStorage()[self.keys])
        DataMiner().key = "first-key"
        self.assertEqual("x", PersistentObjectStorage()[self.keys])

    def test_dict_with_list_data(self):
        DataMiner().data_type = DataTypes.DictWithList
        DataMiner().key = "first-key"
        PersistentObjectStorage().store(keys=self.keys, values="x", metadata={})
        DataMiner().key = "second-key"
        PersistentObjectStorage().store(keys=self.keys, values="y", metadata={})
        PersistentObjectStorage().store(keys=self.keys, values="z", metadata={})

        self.assertIn("first-key", PersistentObjectStorage().storage_object["a"]["b"])
        self.assertIn("second-key", PersistentObjectStorage().storage_object["a"]["b"])

        self.assertEqual("y", PersistentObjectStorage()[self.keys])
        self.assertEqual("z", PersistentObjectStorage()[self.keys])
        DataMiner().key = "first-key"
        self.assertEqual("x", PersistentObjectStorage()[self.keys])


class Metadata(BaseClass):
    keys = ["a", "b"]

    @staticmethod
    @Simple.decorator_plain
    def simple_return(value):
        return value

    def setUp(self):
        super().setUp()
        DataMiner().current_time = time.time()

    def test_latency(self):
        delta = 0.05

        PersistentObjectStorage().store(keys=self.keys, values="x", metadata={})
        self.assertAlmostEqual(
            0, DataMiner().metadata[DataMiner().LATENCY_KEY], delta=delta
        )
        # Try to set some custom metadata
        DataMiner().data.metadata = {"random": "data"}

        time.sleep(0.1)
        PersistentObjectStorage().store(keys=self.keys, values="y", metadata={})
        self.assertAlmostEqual(
            0.1, DataMiner().metadata[DataMiner().LATENCY_KEY], delta=delta
        )

        time.sleep(0.2)
        PersistentObjectStorage().store(keys=self.keys, values="z", metadata={})
        self.assertAlmostEqual(
            0.2, DataMiner().metadata[DataMiner().LATENCY_KEY], delta=delta
        )
        PersistentObjectStorage().read(self.keys)
        self.assertAlmostEqual(
            0, DataMiner().metadata[DataMiner().LATENCY_KEY], delta=delta
        )
        # check custom metadata
        self.assertEqual("data", DataMiner().metadata["random"])

        PersistentObjectStorage().read(self.keys)
        self.assertAlmostEqual(
            0.1, DataMiner().metadata[DataMiner().LATENCY_KEY], delta=delta
        )
        # check if custom metadata are not everywhere
        self.assertNotIn("random", DataMiner().metadata)
        self.assertIn("latency", DataMiner().metadata)
        PersistentObjectStorage().read(self.keys)
        self.assertAlmostEqual(
            0.2, DataMiner().metadata[DataMiner().LATENCY_KEY], delta=delta
        )

    def test_generic(self):
        PersistentObjectStorage().store(
            keys=self.keys, values="x", metadata={"test_meta": "yes"}
        )
        PersistentObjectStorage().metadata = {"rpms": ["package1", "package2"]}
        PersistentObjectStorage().dump()
        PersistentObjectStorage().storage_object = {}
        PersistentObjectStorage().load()
        self.assertIn(self.keys, PersistentObjectStorage())
        self.assertEqual(DataMiner().metadata["test_meta"], "yes")
        self.assertEqual(
            PersistentObjectStorage().metadata.get(
                PersistentObjectStorage().key_inspect_strategy_key
            ),
            StorageKeysInspectDefault.__name__,
        )
        self.assertEqual(
            PersistentObjectStorage().metadata.get("rpms"), ["package1", "package2"]
        )

    def test_strategy(self):
        DataMiner().key_stategy_cls = StorageKeysInspectSimple
        PersistentObjectStorage().store(keys=self.keys, values="x", metadata={})
        self.simple_return("ahoj")
        self.simple_return([1, 2, 3])

        PersistentObjectStorage().dump()
        PersistentObjectStorage().storage_object = {}
        DataMiner().key_stategy_cls = StorageKeysInspectDefault
        PersistentObjectStorage().mode = StorageMode.read
        PersistentObjectStorage().load()
        self.assertEqual("x", PersistentObjectStorage()[self.keys])
        self.assertEqual(
            PersistentObjectStorage().metadata.get(
                PersistentObjectStorage().key_inspect_strategy_key
            ),
            StorageKeysInspectSimple.__name__,
        )
        self.assertEqual(StorageKeysInspectSimple, DataMiner().key_stategy_cls)

    def test_strategy_proper(self):
        self.test_strategy()
        PersistentObjectStorage().load()
        print(PersistentObjectStorage().storage_object)
        self.assertEqual("ahoj", self.simple_return("nonsense"))
        self.assertEqual([1, 2, 3], self.simple_return("nonsense"))

    def test_strategy_API_cls(self):
        self.test_strategy()
        PersistentObjectStorage().load()
        DataMiner().key_stategy_cls = StorageKeysInspect
        self.assertRaises(NotImplementedError, self.simple_return, "nonsense")

    def test_strategy_default_cls(self):
        self.test_strategy()
        PersistentObjectStorage().load()
        DataMiner().key_stategy_cls = StorageKeysInspectDefault
        self.assertEqual("ahoj", self.simple_return("nonsense"))

    def test_strategy_unique(self):
        keys = self.keys + ["c"] + self.keys + ["d"]
        self.assertEqual(
            StorageKeysInspectUnique._get_unique_keys(keys), ["c"] + self.keys + ["d"]
        )


class Latency(BaseClass):
    keys = ["a", "b"]

    def setUp(self):
        super().setUp()
        DataMiner().current_time = time.time()
        DataMiner().use_latency = True

    def tearDown(self):
        DataMiner().use_latency = False

    def test_not_applied(self):
        DataMiner().use_latency = False
        delta = 0.05
        PersistentObjectStorage().store(keys=self.keys, values="x", metadata={})
        time.sleep(0.2)
        PersistentObjectStorage().store(keys=self.keys, values="y", metadata={})
        self.assertAlmostEqual(
            0.2, DataMiner().metadata[DataMiner().LATENCY_KEY], delta=delta
        )

        time_begin = time.time()
        self.assertIn(self.keys, PersistentObjectStorage())
        time_end = time.time()
        self.assertAlmostEqual(0, time_end - time_begin, delta=delta)

        time_begin = time.time()
        self.assertIn(self.keys, PersistentObjectStorage())
        time_end = time.time()
        self.assertAlmostEqual(0, time_end - time_begin, delta=delta)

    def test_applied(self):
        delta = 0.05
        PersistentObjectStorage().store(keys=self.keys, values="x", metadata={})
        time.sleep(0.2)
        PersistentObjectStorage().store(keys=self.keys, values="y", metadata={})
        self.assertAlmostEqual(
            0.2, DataMiner().metadata[DataMiner().LATENCY_KEY], delta=delta
        )

        time_begin = time.time()
        PersistentObjectStorage().read(self.keys)
        time_end = time.time()
        self.assertAlmostEqual(0, time_end - time_begin, delta=delta)

        time_begin = time.time()
        PersistentObjectStorage().read(self.keys)
        time_end = time.time()
        self.assertAlmostEqual(0.2, time_end - time_begin, delta=delta)


class KeySkipping(BaseClass):
    """
    Check fault tolerancy of intermediate keys
    """

    keys = ["a", "b", "c", "d"]

    def setUp(self):
        super().setUp()
        PersistentObjectStorage().store(keys=self.keys, values="x", metadata={})

    def tearDown(self):
        DataMiner().read_key_exact = False

    def test_all_match(self):
        output = PersistentObjectStorage().read(self.keys)
        self.assertEqual("x", output)

    def test_key_1_inserted(self):
        output = PersistentObjectStorage().read(self.keys[:1] + ["y"] + self.keys[1:])
        self.assertEqual("x", output)

    def test_key_0_inserted(self):
        output = PersistentObjectStorage().read(["y"] + self.keys)
        self.assertEqual("x", output)

    def test_key_inserted_2_before(self):
        output = PersistentObjectStorage().read(self.keys[:-2] + ["y"] + self.keys[-2:])
        self.assertEqual("x", output)

    def test_key_inserted_last(self):
        self.assertRaises(
            PersistentStorageException,
            PersistentObjectStorage().read,
            self.keys + ["y"],
        )

    def test_key_inserted_1_before(self):
        self.assertRaises(
            PersistentStorageException,
            PersistentObjectStorage().read,
            self.keys[:-1] + ["y"] + self.keys[-1:],
        )

    def test_key_exact_miner(self):
        DataMiner().read_key_exact = True
        output = PersistentObjectStorage().read(self.keys)
        self.assertEqual("x", output)

    def test_key_exact_miner_exception(self):
        DataMiner().read_key_exact = True
        self.assertRaises(
            PersistentStorageException,
            PersistentObjectStorage().read,
            ["y"] + self.keys,
        )



import os
from requre.storage import PersistentObjectStorage
from requre.helpers.requests_response import RequestResponseHandling
import requests

requests.Session.send = RequestResponseHandling.decorator_plain(requests.Session.send)

PersistentObjectStorage().storage_file = "github.yaml"

import github

g = github.Github(os.getenv("TOKEN", "EMPTY"))
print("Count of your repos: ", len(list(g.get_user().get_repos())))

PersistentObjectStorage().dump()



from requre.helpers.files import StoreFiles
from requre.storage import PersistentObjectStorage
from tempfile import mkdtemp
import os

PersistentObjectStorage().storage_file = "files.yaml"
temp_dir = mkdtemp()


@StoreFiles.guess_args
def return_result(dir_name):
    file_name = input("enter file name: ")
    with open(os.path.join(dir_name, file_name), "w") as fd:
        fd.write("empty")
    return file_name


print("returned file name:", return_result(temp_dir))
print("dir name (current):", temp_dir)
print("files:", os.listdir(temp_dir))
PersistentObjectStorage().dump()



import os
import inspect
from unittest import TestCase
from requre.storage import PersistentObjectStorage
from requre.constants import RELATIVE_TEST_DATA_DIRECTORY


class RequreTestCase(TestCase):
    """
    This is base unittest TestCase what will help you to store requre storage files
    It will create directory structre test_data/TEST_FILE_NAME in a directory where
    test is stored. inside this subdir. there is created storage file via calling
    self.id()
    suffixed by yaml
    """

    def get_datafile_filename(self, suffix="yaml"):
        current_class_file = inspect.getfile(self.__class__)
        real_path_dir = os.path.realpath(os.path.dirname(current_class_file))
        test_file_name = os.path.basename(current_class_file).rsplit(".", 1)[0]
        test_name = self.id()
        testdata_dirname = os.path.join(
            real_path_dir, RELATIVE_TEST_DATA_DIRECTORY, test_file_name
        )
        return os.path.join(testdata_dirname, f"{test_name}.{suffix}")

    def setUp(self):
        super().setUp()
        response_file = self.get_datafile_filename()
        os.makedirs(os.path.dirname(response_file), mode=0o777, exist_ok=True)
        PersistentObjectStorage().storage_file = response_file

    def tearDown(self):
        PersistentObjectStorage().dump()
        super().tearDown()



# MIT License
#
# Copyright (c) 2018-2019 Red Hat, Inc.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import logging
from typing import Optional
from requre.storage import StorageCounter

logger = logging.getLogger(__name__)


class TempFile(StorageCounter):
    """
    replace system tempfile module with own predictable names implementation
     of temp files for mocking
    """

    root = "/tmp"
    prefix = "static_tmp"

    @classmethod
    def _get_name(cls, prefix: Optional[str] = None) -> str:
        cls.reset_counter_if_changed()
        filename = f"{prefix or cls.prefix}_{cls.next()}"
        os.makedirs(
            os.path.join(cls.root, cls.storage_file()), mode=0o777, exist_ok=True
        )
        output = os.path.join(cls.root, cls.storage_file(), filename)
        logger.debug(f"Use name for tempfile: ${output}")
        return output

    @classmethod
    def mktemp(cls, prefix: Optional[str] = None) -> str:
        return cls._get_name(prefix)

    @classmethod
    def mkdtemp(cls, prefix: Optional[str] = None) -> str:
        name = cls._get_name(prefix)
        os.makedirs(name)
        return name



import os
import tarfile
import tempfile
from io import BytesIO

from requre.helpers.files import StoreFiles
from requre.storage import PersistentObjectStorage
from requre.utils import StorageMode
from tests.test_function_output import StoreFunctionOutput
from tests.testbase import BaseClass


class Base(BaseClass):
    @StoreFiles.arg_references({"target_file": 2})
    def create_file_content(self, value, target_file):
        with open(target_file, "w") as fd:
            fd.write(value)
        return "value"

    @StoreFiles.arg_references({"target_dir": 2})
    def create_dir_content(self, filename, target_dir, content="empty"):
        with open(os.path.join(target_dir, filename), "w") as fd:
            fd.write(content)


class FileStorage(Base):
    def test_base(self):
        """
        verify if tar archive is in persistent storage
        """
        filename = "some_random_name"
        self.create_temp_dir()
        self.create_dir_content(
            filename=filename, target_dir=self.temp_dir, content="ciao"
        )
        tar_data = PersistentObjectStorage().storage_object["X"]["file"]["tar"][
            "StoreFiles"
        ]["storage_test.yaml"]["target_dir"][0]["output"]
        with BytesIO(tar_data) as tar_byte:
            with tarfile.open(mode="r:xz", fileobj=tar_byte) as tar_archive:
                self.assertIn(filename, [x.name for x in tar_archive.getmembers()][-1])

    def test_create_file_content(self):
        """
        test if is able store files via decorated function create_file_content
        """
        self.create_temp_file()
        self.assertEqual(
            "value", self.create_file_content("ahoj", target_file=self.temp_file)
        )

        self.create_file_content("cao", target_file=self.temp_file)

        PersistentObjectStorage().dump()
        PersistentObjectStorage().mode = StorageMode.read

        self.create_file_content("first", target_file=self.temp_file)
        with open(self.temp_file, "r") as fd:
            content = fd.read()
            self.assertIn("ahoj", content)
            self.assertNotIn("cao", content)
        self.create_file_content("second", target_file=self.temp_file)
        with open(self.temp_file, "r") as fd:
            content = fd.read()
            self.assertNotIn("ahoj", content)
            self.assertIn("cao", content)
        self.assertRaises(
            Exception, self.create_file_content, "third", target_file=self.temp_file
        )

    def test_create_file_content_positional(self):
        """
        Similar to  test_create_file_content,
        but test it also via positional parameters and mixing them
        """
        self.create_temp_file()
        self.create_file_content("ahoj", self.temp_file)
        self.create_file_content("cao", self.temp_file)

        PersistentObjectStorage().dump()
        PersistentObjectStorage().mode = StorageMode.read

        self.create_temp_file()
        self.create_file_content("first", self.temp_file)
        with open(self.temp_file, "r") as fd:
            content = fd.read()
            self.assertIn("ahoj", content)
            self.assertNotIn("cao", content)
        # mix with positional option

        self.create_temp_file()
        self.create_file_content("second", target_file=self.temp_file)
        with open(self.temp_file, "r") as fd:
            content = fd.read()
            self.assertNotIn("ahoj", content)
            self.assertIn("cao", content)
        self.create_temp_file()
        self.assertRaises(
            Exception, self.create_file_content, "third", target_file=self.temp_file
        )

    def test_create_dir_content(self):
        """
        Check if properly store and restore directory content
        """
        self.create_temp_dir()
        self.create_dir_content(
            filename="ahoj", target_dir=self.temp_dir, content="ciao"
        )
        self.assertIn("ahoj", os.listdir(self.temp_dir))
        with open(os.path.join(self.temp_dir, "ahoj"), "r") as fd:
            content = fd.read()
            self.assertIn("ciao", content)

        PersistentObjectStorage().dump()
        PersistentObjectStorage().mode = StorageMode.read

        self.create_temp_dir()
        self.create_dir_content(
            filename="nonsense", target_dir=self.temp_dir, content="bad"
        )
        self.assertIn("ahoj", os.listdir(self.temp_dir))
        self.assertNotIn("nonsense", os.listdir(self.temp_dir))
        with open(os.path.join(self.temp_dir, "ahoj"), "r") as fd:
            content = fd.read()
            self.assertIn("ciao", content)
            self.assertNotIn("bad", content)

    def test_hidden_file(self):
        """
        Check if properly store and restore directory content with hidden file
        """
        filename = ".git"
        self.create_temp_dir()
        self.create_dir_content(
            filename=filename, target_dir=self.temp_dir, content="hidden"
        )

        PersistentObjectStorage().dump()
        PersistentObjectStorage().mode = StorageMode.read

        self.create_temp_dir()
        self.create_dir_content(
            filename="nonsense", target_dir=self.temp_dir, content="bad"
        )
        self.assertIn(filename, os.listdir(self.temp_dir))
        self.assertNotIn("nonsense", os.listdir(self.temp_dir))
        with open(os.path.join(self.temp_dir, filename), "r") as fd:
            content = fd.read()
            self.assertIn("hidden", content)
            self.assertNotIn("bad", content)


class SessionRecordingWithFileStore(Base):
    @StoreFiles.arg_references({"target_file": 2})
    def create_file_content(self, value, target_file):
        StoreFunctionOutput.run_command_wrapper(
            cmd=["bash", "-c", f"echo {value} > {target_file}"]
        )
        with open(target_file, "w") as fd:
            fd.write(value)

    def test(self):
        """
        Mixing command wrapper with file storage
        """
        self.create_temp_file()
        self.create_file_content("ahoj", target_file=self.temp_file)
        self.create_file_content("cao", target_file=self.temp_file)

        PersistentObjectStorage().dump()
        PersistentObjectStorage().mode = StorageMode.read

        before = str(PersistentObjectStorage().storage_object)

        self.create_file_content("ahoj", target_file=self.temp_file)
        with open(self.temp_file, "r") as fd:
            content = fd.read()
            self.assertIn("ahoj", content)
            self.assertNotIn("cao", content)
        self.create_file_content("cao", target_file=self.temp_file)
        with open(self.temp_file, "r") as fd:
            content = fd.read()
            self.assertNotIn("ahoj", content)
            self.assertIn("cao", content)
        after = str(PersistentObjectStorage().storage_object)
        self.assertGreater(len(before), len(after))
        self.assertIn("True", before)


class DynamicFileStorage(Base):
    @StoreFiles.guess_args
    def write_to_file(self, value, target_file):
        with open(target_file, "w") as fd:
            fd.write(value)

    def test_create_file(self):
        """
        File storage where it try to guess what to store, based on *args and **kwargs values
        """
        self.create_temp_file()
        self.write_to_file("ahoj", self.temp_file)
        self.write_to_file("cao", self.temp_file)

        PersistentObjectStorage().dump()
        PersistentObjectStorage().mode = StorageMode.read
        self.create_temp_file()
        self.write_to_file("ahoj", self.temp_file)
        with open(self.temp_file, "r") as fd:
            content = fd.read()
            self.assertIn("ahoj", content)
            self.assertNotIn("cao", content)
        # mix with positional option

        self.create_temp_file()
        self.write_to_file("cao", self.temp_file)
        with open(self.temp_file, "r") as fd:
            content = fd.read()
            self.assertNotIn("ahoj", content)
            self.assertIn("cao", content)
        self.create_temp_file()
        self.assertRaises(
            Exception, self.create_file_content, "third", target_file=self.temp_file
        )


class StoreOutputFile(Base):
    @StoreFiles.return_value
    def create_file(self, value):
        tmpfile = tempfile.mktemp()
        with open(tmpfile, "w") as fd:
            fd.write(value)
        return tmpfile

    def test_create_file(self):
        """
        Test File storage if file name is return value of function
        """
        ofile1 = self.create_file("ahoj")
        ofile2 = self.create_file("cao")

        PersistentObjectStorage().dump()
        PersistentObjectStorage().mode = StorageMode.read

        oofile1 = self.create_file("first")
        with open(ofile1, "r") as fd:
            content = fd.read()
            self.assertIn("ahoj", content)
            self.assertNotIn("cao", content)
        with open(oofile1, "r") as fd:
            content = fd.read()
            self.assertIn("ahoj", content)
            self.assertNotIn("cao", content)
        # mix with positional option

        oofile2 = self.create_file("second")
        with open(oofile2, "r") as fd:
            content = fd.read()
            self.assertNotIn("ahoj", content)
            self.assertIn("cao", content)
        self.assertEqual(ofile2, oofile2)



import os
import shutil
import tempfile
import unittest

from requre.storage import PersistentObjectStorage
from requre.utils import StorageMode


class BaseClass(unittest.TestCase):
    def setUp(self) -> None:
        PersistentObjectStorage().mode = StorageMode.default
        super().setUp()
        self.file_name = None
        self.temp_dir = None
        self.temp_file = None
        self.response_dir = tempfile.mkdtemp(prefix="data_store")
        self.response_file = os.path.join(self.response_dir, "storage_test.yaml")
        PersistentObjectStorage().storage_file = self.response_file
        PersistentObjectStorage().dump_after_store = True

    def tearDown(self) -> None:
        super().tearDown()
        shutil.rmtree(self.response_dir)
        if self.file_name:
            os.remove(self.file_name)
        if self.temp_dir and os.path.exists(self.temp_dir):
            shutil.rmtree(self.temp_dir)
        if self.temp_file and os.path.exists(self.temp_file):
            os.remove(self.temp_file)

    def create_temp_dir(self):
        if self.temp_dir:
            shutil.rmtree(self.temp_dir)
        self.temp_dir = tempfile.mkdtemp()

    def create_temp_file(self):
        if self.temp_file:
            os.remove(self.temp_file)
        self.temp_file = tempfile.mktemp()



#!/usr/bin/python3

import os
import sys
import shutil
import click
import importlib.util
import atexit
from typing import Any
import yaml
import builtins

from requre.import_system import upgrade_import_system, UpgradeImportSystem
from requre.postprocessing import DictProcessing
from requre.storage import DataMiner, PersistentObjectStorage
from requre.constants import (
    ENV_REPLACEMENT_FILE,
    ENV_STORAGE_FILE,
    ENV_DEBUG,
    REPLACE_DEFAULT_KEY,
    ENV_APPLY_LATENCY,
    ENV_REPLACEMENT_NAME,
)

"""
This is command line tool for E2E and functional testing to enable requre
without access python code.

It apply itself as sitecustomize.py script
to user home path:
   ~/.local/lib/python{version}/site-packages
or to system path via option --system:
   /usr/lib/python{version}/site-packages

when tool is installed call python code with enviroment variables:
 RESPONSE_FILE - Storage file path for session recording.
        In case file does not exists, it will use write mode
        In case file exists, it will use read mode for Storage
 REPLACEMENT_FILE - Replacement file  path for import system.
        It is important to have there set variable FILTERS what will
        be used as replacements list for upgrade_import_system function.
        For more details see doc: https://github.com/packit-service/requre/
 REPLACEMENT_VAR - Overrides default value of variable in REPLACEMENT_FILE
        what will be used as replacement variable.
 DEBUG - if set, print debugging information, fi requre is applied
 LATENCY - apply latency waits for test, to have simiar test timing
        It is important when using some async/messaging calls
"""

FILE_NAME = "sitecustomize.py"


def debug_print(*args):
    if os.getenv(ENV_DEBUG):
        print("REQURE DEBUG:", *args, file=sys.__stderr__)


def raise_error(ret_code: int, msg: Any):
    """
    When installed as sitecustomization.py, exceptions are not propagated to main process
    process, ends successfully, although it contains traceback/

    :param ret_code: return code to return
    :param msg: message to write to stderr
    :return: None
    """
    print(msg)
    os._exit(ret_code)


def apply_fn():
    """
    This function is used when installed as  sitecustomize.py script
    to enable replacing system, please set env vars RESPONSE_FILE
    REPLACEMENT_FILE, see doc string of this file
    """
    # file name of storage file
    storage_file = os.getenv(ENV_STORAGE_FILE)
    # file name of replaces for updated import system
    replacement_file = os.getenv(ENV_REPLACEMENT_FILE)
    if_latency = os.getenv(ENV_APPLY_LATENCY)
    replacement_var = os.getenv(ENV_REPLACEMENT_NAME, REPLACE_DEFAULT_KEY)
    debug_print(
        f"You have patched version of your python by requre project "
        f"(python {sys.version_info.major}.{sys.version_info.minor}, {__file__}) "
    )
    if not (storage_file and replacement_file):
        debug_print(
            f"\tYou have to set {ENV_STORAGE_FILE} and "
            f"{ENV_REPLACEMENT_FILE} env variables to work properly"
        )
    else:
        if not os.path.exists(replacement_file):
            raise FileExistsError(
                f"{replacement_file} has to exist to work properly "
                f"(python file with replacements definition)"
            )
        if if_latency:
            debug_print("Use latency for function calls")
            DataMiner().use_latency = True
        PersistentObjectStorage().storage_file = storage_file
        spec = importlib.util.spec_from_file_location("replacements", replacement_file)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        if hasattr(module, replacement_var):
            replacement = getattr(module, replacement_var)
            debug_print(f"Replaces: {replacement}")
            if isinstance(replacement, UpgradeImportSystem):
                debug_print(
                    f"{replacement_var} is {UpgradeImportSystem.__name__} object"
                )
            elif isinstance(replacement, list):
                debug_print(
                    f"{replacement_var} is list of replacements, apply upgrading"
                )
                upgrade_import_system(filters=replacement)
            else:
                raise_error(126, f"Bad type of {replacement_var}, see documentation")
        else:
            raise_error(
                125,
                f"in {replacement_file} there is not defined '{replacement_var}' variable",
            )
        # register dump command, when python finish
        atexit.register(PersistentObjectStorage().dump)


def get_current_python_version():
    """
    Get current running version of python

    :return: str in format X.Y
    """
    return f"{sys.version_info.major}.{sys.version_info.minor}"


def path_to_python_customize(version: str, global_path: bool = False):
    """
    return path to sitecustomize.py file based on python version and
    if it should be local for user or installed to system

    :param version: str: version string
    :param global_path: bool - if apply it to whole system
    :return: str with full path to python file
    """
    if global_path:
        site_path = f"/usr/lib/python{version}/site-packages"
    else:
        site_path = os.path.expanduser(f"~/.local/lib/python{version}/site-packages")
        os.makedirs(site_path, exist_ok=True)
    customize_file = os.path.join(site_path, FILE_NAME)
    return customize_file


def patch_verify(pathname: str):
    """
    Check if patch file already exists
    :param pathname: path to sitecustomization file
    :return: bool if exists
    """
    return os.path.exists(pathname)


@click.group("requre")
@click.option(
    "--version", default=get_current_python_version(), help="Version of python to patch"
)
@click.option(
    "--system",
    is_flag=True,
    default=False,
    help="Use system python path, instead of user home dir",
)
@click.pass_context
def requre_base(ctx, version, system):
    ctx.obj = {FILE_NAME: path_to_python_customize(version=version, global_path=system)}


@requre_base.command()
@click.pass_context
def verify(ctx):
    if patch_verify(ctx.obj[FILE_NAME]):
        click.echo(f"Python patched (file: {ctx.obj[FILE_NAME]})")
    else:
        raise click.ClickException(f"Python not patched (file: {ctx.obj[FILE_NAME]})")


@requre_base.command()
@click.pass_context
def apply(ctx):
    if patch_verify(ctx.obj[FILE_NAME]):
        raise click.ClickException(
            f"Python already patched (file: {ctx.obj[FILE_NAME]})"
        )
    else:
        click.echo(f"Applying import patch to python (file: {ctx.obj[FILE_NAME]})")
        shutil.copy(__file__, ctx.obj[FILE_NAME])


@requre_base.command()
@click.pass_context
def clean(ctx):
    if patch_verify(ctx.obj[FILE_NAME]):
        os.remove(ctx.obj[FILE_NAME])
    else:
        raise click.ClickException(
            f"Patch not applied (file: {ctx.obj[FILE_NAME]}), nothing to do"
        )


@requre_base.command()
@click.option(
    "--replaces",
    help="match_string:key:type_of_value:value = Substitution query in format, "
    "where match_string is in format of selecting dictionary keys:"
    "selector1%selector2, type_of_value is some object what is serializable "
    "and part or builtins module (e.g. int)",
    multiple=True,
)
@click.argument("files", nargs=-1, type=click.File("r"))
@click.option(
    "--dry-run", is_flag=True, default=False, help="Do not write changes back"
)
@click.option(
    "--simplify",
    is_flag=True,
    default=False,
    help="Simplify dict structure if possible (experimental feature)",
)
def purge(replaces, files, dry_run, simplify):
    for one_file in files:
        click.echo(f"Processing file: {one_file.name}")
        object_representation = yaml.safe_load(one_file)
        processor = DictProcessing(object_representation)
        for item in replaces:
            click.echo(f"\tTry to apply: {item}")
            selector_str, key, type_of_value, value = item.split(":", 3)
            selector_list = [] if not selector_str else selector_str.split("%")
            # retype the output object to proper type ()
            value = getattr(builtins, type_of_value)(value)
            for matched in processor.match(selector=selector_list):
                click.echo(f"\t\tMatched {selector_list}")
                processor.replace(obj=matched, key=key, value=value)
        if simplify:
            processor.simplify()
        if not dry_run:
            click.echo(f"Writing content back to file: {one_file.name}")
            with open(one_file.name, mode="w") as outfile:
                outfile.write(yaml.safe_dump(object_representation))


if __name__ == "__main__" or not (__file__ and __file__.endswith(FILE_NAME)):
    requre_base()
else:
    apply_fn()



from requre.import_system import decorate, replace, replace_module, log_imports
from requre.base_testclass import RequreTestCase

__all__ = [
    decorate.__name__,
    replace.__name__,
    replace_module.__name__,
    log_imports.__name__,
    RequreTestCase.__name__,
]



import tempfile
from requre.storage import PersistentObjectStorage

keys = ["basic", "keys", 0.1]
value1 = {"cat": "tom", "mouse": "jerry"}
value2 = {"cat1": "pa", "cat2": "pi"}

PersistentObjectStorage().storage_file = tempfile.mktemp()

PersistentObjectStorage().store(keys, value1, metadata={})
PersistentObjectStorage().store(keys, value2, metadata={})
PersistentObjectStorage().dump()

PersistentObjectStorage().storage_object = {}
PersistentObjectStorage().load()
print(PersistentObjectStorage().storage_object)
print(PersistentObjectStorage().read(keys))
print(PersistentObjectStorage().read(keys))



import os

from requre.helpers.tempfile import TempFile
from requre.storage import PersistentObjectStorage
from tests.testbase import BaseClass


class TestTempFile(BaseClass):
    def testSimple(self):
        output = TempFile.mktemp()
        self.assertIn(
            f"/tmp/{os.path.basename(PersistentObjectStorage().storage_file)}/static_tmp_1",
            output,
        )

    def testChangeFile(self):
        PersistentObjectStorage().storage_file += ".x"
        output = TempFile.mktemp()
        self.assertIn(
            f"/tmp/{os.path.basename(PersistentObjectStorage().storage_file)}/static_tmp_1",
            output,
        )
        output = TempFile.mktemp()
        self.assertIn(
            f"/tmp/{os.path.basename(PersistentObjectStorage().storage_file)}/static_tmp_2",
            output,
        )
        PersistentObjectStorage().storage_file += ".y"
        self.assertEqual(TempFile.counter, 2)
        output = TempFile.mktemp()
        self.assertEqual(TempFile.counter, 1)
        self.assertIn(
            f"/tmp/{os.path.basename(PersistentObjectStorage().storage_file)}/static_tmp_1",
            output,
        )
        output = TempFile.mktemp()
        self.assertIn(
            f"/tmp/{os.path.basename(PersistentObjectStorage().storage_file)}/static_tmp_2",
            output,
        )



VERSION_REQURE_FILE = 2
ENV_STORAGE_FILE = "RESPONSE_FILE"
ENV_REPLACEMENT_FILE = "REPLACEMENT_FILE"
ENV_REPLACEMENT_NAME = "REPLACEMENT_VAR"
ENV_DEBUG = "DEBUG"
ENV_APPLY_LATENCY = "LATENCY"
ENV_REQURE_STORAGE_MODE = "REQURE_MODE"
REPLACE_DEFAULT_KEY = "FILTERS"
METATADA_KEY = "_requre"
KEY_MINIMAL_MATCH = 2
RELATIVE_TEST_DATA_DIRECTORY = "test_data"



import os
import tempfile
import shutil
import yaml
from unittest import TestCase
from requre.constants import ENV_REQURE_STORAGE_MODE
from requre.exceptions import PersistentStorageException
from requre.storage import PersistentObjectStorage, StorageMode


class GenericModeTests(TestCase):
    def tearDown(self) -> None:
        if ENV_REQURE_STORAGE_MODE in os.environ:
            del os.environ[ENV_REQURE_STORAGE_MODE]
        super().tearDown()

    def test_bad_mode(self):
        os.environ[ENV_REQURE_STORAGE_MODE] = "mistake"
        with self.assertRaises(PersistentStorageException) as context:
            PersistentObjectStorage().storage_file = "/not/important"
            self.assertTrue(
                f"storage mode mistake does not exist," in context.exception
            )

    def test_default(self):
        PersistentObjectStorage()._set_defaults()
        self.assertEqual(PersistentObjectStorage().mode, StorageMode.default)


class ReadMode(TestCase):
    def setUp(self) -> None:
        super().setUp()
        os.environ[ENV_REQURE_STORAGE_MODE] = "read"
        self.response_dir = tempfile.mkdtemp(prefix="data_store")
        self.response_file = os.path.join(self.response_dir, "storage_test.yaml")

    def tearDown(self) -> None:
        shutil.rmtree(self.response_dir)
        if ENV_REQURE_STORAGE_MODE in os.environ:
            del os.environ[ENV_REQURE_STORAGE_MODE]
        super().tearDown()

    def test_existing_file(self):
        with open(self.response_file, "w") as fd:
            yaml.dump({}, fd)
        PersistentObjectStorage().storage_file = self.response_file
        self.assertEqual(PersistentObjectStorage().mode, StorageMode.read)

    def test_non_existing_file(self):
        with self.assertRaises(PersistentStorageException) as context:
            PersistentObjectStorage().storage_file = self.response_file
            self.assertTrue("does not exist" in context.exception)


class WriteMode(TestCase):
    def setUp(self) -> None:
        super().setUp()
        os.environ[ENV_REQURE_STORAGE_MODE] = "write"
        self.response_dir = tempfile.mkdtemp(prefix="data_store")
        self.response_file = os.path.join(self.response_dir, "storage_test.yaml")

    def tearDown(self) -> None:
        shutil.rmtree(self.response_dir)
        if ENV_REQURE_STORAGE_MODE in os.environ:
            del os.environ[ENV_REQURE_STORAGE_MODE]
        super().tearDown()

    def test_existing_file(self):
        with open(self.response_file, "w") as fd:
            yaml.dump({}, fd)
        PersistentObjectStorage().storage_file = self.response_file
        self.assertEqual(PersistentObjectStorage().mode, StorageMode.write)

    def test_non_existing_file(self):
        PersistentObjectStorage().storage_file = self.response_file


class AppendMode(WriteMode):
    def setUp(self) -> None:
        super().setUp()
        os.environ[ENV_REQURE_STORAGE_MODE] = "append"

    def test_existing_file(self):
        with open(self.response_file, "w") as fd:
            yaml.dump({}, fd)
        PersistentObjectStorage().storage_file = self.response_file
        self.assertEqual(PersistentObjectStorage().mode, StorageMode.append)



import os
from requre.storage import PersistentObjectStorage
from requre.helpers.requests_response import RequestResponseHandling
from requre.import_system import upgrade_import_system

upgrade_import_system().decorate(
    where="^requests$",
    what="Session.send",
    who_name=["github"],
    decorator=RequestResponseHandling.decorator_plain,
)

PersistentObjectStorage().storage_file = "github2.yaml"

import github

g = github.Github(os.getenv("TOKEN", "EMPTY"))
print("Count of your repos: ", len(list(g.get_user().get_repos())))

PersistentObjectStorage().dump()



import logging
from typing import Union, Any, Dict, Optional, List
from .constants import KEY_MINIMAL_MATCH, METATADA_KEY
from .storage import DataMiner, DataStructure, DataTypes

logger = logging.getLogger(__name__)


class DictProcessing:
    def __init__(self, requre_dict: dict):
        self.requre_dict = requre_dict

    def match(self, selector: list, internal_object: Union[dict, list, None] = None):
        if internal_object is None:
            internal_object = self.requre_dict
        if len(selector) == 0:
            logger.debug(f"all selectors matched")
            yield internal_object
            # add return here, to avoid multiple returns
            return
        if isinstance(internal_object, dict):
            for k, v in internal_object.items():
                if v is None:
                    return
                if selector and selector[0] == k:
                    logger.debug(f"selector {k} matched")
                    yield from self.match(selector=selector[1:], internal_object=v)
                else:
                    yield from self.match(selector=selector, internal_object=v)
        elif isinstance(internal_object, list):
            for list_item in internal_object:
                if list_item is None:
                    return
                yield from self.match(selector=selector, internal_object=list_item)
        else:
            return

    @staticmethod
    def replace(obj: Any, key: Any, value: Any) -> None:
        if isinstance(obj, dict):
            for k, v in obj.items():
                if k == key:
                    logger.debug(f"replacing: {obj[key]} by {value}")
                    obj[key] = value
                else:
                    DictProcessing.replace(obj=v, key=key, value=value)
        if isinstance(obj, list):
            for item in obj:
                DictProcessing.replace(obj=item, key=key, value=value)

    @staticmethod
    def minimal_match(dict_obj: Dict, metadata: Dict):

        tmp_dict = dict_obj
        first_item: Dict = {}
        key_name = DataTypes.__name__
        ds = DataTypes.List
        if key_name in metadata:
            ds = DataTypes(metadata.get(key_name))
        logger.debug(f"Use datatype: {ds}")
        for cntr in range(KEY_MINIMAL_MATCH):
            if not isinstance(tmp_dict, dict) or len(tmp_dict.keys()) != 1:
                return False
            key = list(tmp_dict.keys())[0]
            value = tmp_dict[key]
            tmp_dict = value
        if ds == DataTypes.DictWithList:
            if isinstance(tmp_dict, dict):
                tmp_first_item = tmp_dict[list(tmp_dict.keys())[0]]
                if isinstance(tmp_first_item, list):
                    first_item = tmp_first_item[0]
        if ds == DataTypes.Dict:
            if isinstance(tmp_dict, dict):
                first_item = tmp_dict[list(tmp_dict.keys())[0]]
        if ds == DataTypes.List:
            if isinstance(tmp_dict, list):
                first_item = tmp_dict[0]
        if ds == DataTypes.Value:
            if isinstance(tmp_dict, dict):
                first_item = tmp_dict
        if isinstance(first_item, dict) and DataMiner().LATENCY_KEY in first_item.get(
            DataStructure.METADATA_KEY, {}
        ):
            logger.debug(
                f"Minimal key path ({KEY_MINIMAL_MATCH}) matched for {first_item.keys()}"
            )
            return True
        return False

    def simplify(
        self, internal_object: Optional[Dict] = None, ignore_list: Optional[List] = None
    ):
        ignore_list = ignore_list or []
        internal_object = (
            self.requre_dict if internal_object is None else internal_object
        )
        if isinstance(internal_object, dict):
            if len(internal_object.keys()) == 1:
                key = list(internal_object.keys())[0]
                if key in [METATADA_KEY] + ignore_list:
                    return
                if self.minimal_match(internal_object, self.requre_dict[METATADA_KEY]):
                    return
                if isinstance(internal_object[key], dict):
                    value = internal_object.pop(key)
                    print(f"Removing key: {key}  -> {list(value.keys())}")
                    for k, v in value.items():
                        internal_object[k] = v
                        self.simplify(
                            internal_object=internal_object, ignore_list=ignore_list
                        )
            else:
                for v in internal_object.values():
                    self.simplify(internal_object=v, ignore_list=ignore_list)



# MIT License
#
# Copyright (c) 2018-2019 Red Hat, Inc.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import functools
import logging
import pickle
from typing import Optional, Callable, Any, List, Dict

from requre.storage import (
    PersistentObjectStorage,
    DataMiner,
    original_time,
    StorageKeysInspectFull,
)

logger = logging.getLogger(__name__)


class ObjectStorage:
    """
    Generic object API for objects for persistent storage.
    Use it as parent class for your custom object handling

    This class use pickle module for object serialization,
    it is generic, but may lead to some issues
    in case object is not serializable well. Use this class very carefully.
    """

    persistent_storage = PersistentObjectStorage()
    __response_keys: list = list()
    object_type = object

    def __init__(
        self, store_keys: list, pstorage: Optional[PersistentObjectStorage] = None
    ) -> None:
        self.store_keys = store_keys
        if pstorage:
            self.persistent_storage = pstorage
        self.store_keys = store_keys

    @staticmethod
    def get_base_keys(func: Callable) -> List[Any]:
        return DataMiner().key_stategy_cls.get_base_keys(func)

    @classmethod
    def execute(cls, keys: list, func: Callable, *args, **kwargs) -> Any:
        """
        Class method to store or read object from persistent storage
        with explicit set of *args, **kwargs parameters to use as keys
        :param keys: list of keys used as parameter
        :param func: original function
        :param args: parameters of original function
        :param kwargs: parameters of original function
        :return: output of called func
        """

        object_storage = cls(store_keys=keys)

        if object_storage.persistent_storage.do_store(keys):
            time_before = original_time()
            response = func(*args, **kwargs)
            time_after = original_time()
            metadata: Dict = {
                DataMiner().LATENCY_KEY: time_after - time_before,
                DataMiner().METADATA_CALLER_LIST: StorageKeysInspectFull.get_base_keys(
                    func
                ),
            }
            if DataMiner().store_arg_debug_metadata:
                args_clean = [f"'{x}'" if isinstance(x, str) else str(x) for x in args]
                kwargs_clean = [
                    f"""{k}={f"'{v}'" if isinstance(v, str) else str(v)}"""
                    for k, v in kwargs.items()
                ]
                caller = f"{func.__name__}({', '.join(args_clean + kwargs_clean)})"
                metadata[DataMiner().METADATA_ARG_DEBUG_KEY] = caller
            object_storage.write(response, metadata)
            logger.debug(f"WRITE Keys: {keys} -> {response}")
            return response

        else:
            response = object_storage.read()
            logger.debug(f"READ  Keys: {keys} -> {response}")
            return response

    @classmethod
    def execute_all_keys(cls, func: Callable, *args, **kwargs):
        """
        Class method what does same as execute, but use all *args, **kwargs as keys
        :param func: original function
        :param args: parameters of original function
        :param kwargs: parameters of original function
        :return: output of called func
        """
        keys = (
            cls.get_base_keys(func)
            + [x for x in args if isinstance(int, str)]
            + [f"{k}:{v}" for k, v in kwargs.items()]
        )
        return cls.execute(keys, func, *args, **kwargs)

    @classmethod
    def execute_plain(cls, func: Callable, *args, **kwargs):
        """
        Class method what does same as execute, but use just name of module and function as name

        :param func: original function
        :param args: parameters of original function
        :param kwargs: parameters of original function
        :return: output of called func
        """
        keys = cls.get_base_keys(func)
        return cls.execute(keys, func, *args, **kwargs)

    @classmethod
    def decorator_all_keys(cls, func: Callable) -> Any:
        """
        Class method for what should be used as decorator of import replacing system
        This use all arguments of function as keys

        :param func: Callable object
        :return: output of func
        """

        @functools.wraps(func)
        def internal(*args, **kwargs):
            return cls.execute_all_keys(func, *args, **kwargs)

        return internal

    @classmethod
    def decorator(cls, *, item_list: list) -> Any:
        """
        Class method for what should be used as decorator of import replacing system
        This use list of selection of *args or **kwargs as arguments of function as keys

        :param item_list: list of values of *args nums,  **kwargs names to use as keys
        :return: output of func
        """

        def internal(func: Callable):
            @functools.wraps(func)
            def internal_internal(*args, **kwargs):
                keys = cls.get_base_keys(func)
                for item in item_list:
                    if isinstance(item, int):
                        keys.append(args[item])
                    else:
                        keys.append(kwargs[item])
                return cls.execute(keys, func, *args, **kwargs)

            return internal_internal

        return internal

    @classmethod
    def decorator_plain(cls, func: Callable) -> Any:
        """
        Class method for what should be used as decorator of import replacing system
        This use no arguments of function as keys

        :param func: Callable object
        :return: output of func
        """

        @functools.wraps(func)
        def internal(*args, **kwargs):
            return cls.execute_plain(func, *args, **kwargs)

        return internal

    def write(self, obj: Any, metadata: Optional[Dict] = None) -> Any:
        """
        Write the object representation to storage
        Internally it will use self.to_serializable()
        method to get serializable object representation

        :param obj: some object
        :param metadata: store metedata to object
        :return: same obj
        """
        self.persistent_storage.store(
            self.store_keys, self.to_serializable(obj), metadata=metadata
        )
        return obj

    def read(self):
        """
        Crete object representation of serialized data in persistent storage
        Internally it will use self.from_serializable method transform object

        :return: proper object
        """
        data = self.persistent_storage[self.store_keys]
        obj = self.from_serializable(data)
        return obj

    def to_serializable(self, obj: Any) -> Any:
        """
        Internal method for object serialization

        :param obj: some object
        :return: Yaml serializable object
        """
        output = pickle.dumps(obj)
        return output

    def from_serializable(self, data: Any) -> Any:
        """
        Internal method for object de-serialization

        :param data: Yaml serializable object
        :return: some object
        """
        output = pickle.loads(data)
        return output



# MIT License
#
# Copyright (c) 2018-2019 Red Hat, Inc.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import datetime
import json
from io import BytesIO
from typing import Any, Optional, Dict

from requests.models import Response
from requests.structures import CaseInsensitiveDict

from requre.objects import ObjectStorage


class RequestResponseHandling(ObjectStorage):
    __response_keys = ["status_code", "encoding", "reason"]
    __ignored = ["cookies"]
    __response_keys_special = ["raw", "_next", "headers", "elapsed", "_content"]
    __store_indicator = "__store_indicator"
    __implicit_encoding = "UTF-8"

    def write(self, response: Response, metadata: Optional[Dict] = None) -> Response:
        super().write(response, metadata)
        if getattr(response, "next"):
            self.write(getattr(response, "next"))
        return response

    def read(self):
        data = super().read()
        if getattr(data, "next"):
            data._next = self.read()
        return data

    def to_serializable(self, response: Response) -> Any:
        output = dict()
        for key in self.__response_keys:
            output[key] = getattr(response, key)
        for key in self.__response_keys_special:
            if key == "raw":
                binary_data = response.raw.read()
                output[key] = binary_data
                # replay it back to raw
                response.raw = BytesIO(binary_data)
            if key == "headers":
                output[key] = dict(response.headers)
            if key == "elapsed":
                output[key] = response.elapsed.total_seconds()
            if key == "_content":
                what_store = response._content  # type: ignore
                encoding = response.encoding or self.__implicit_encoding
                try:
                    what_store = what_store.decode(encoding)
                    try:
                        what_store = json.loads(what_store)
                        indicator = 2
                    except json.decoder.JSONDecodeError:
                        indicator = 1
                except (ValueError, AttributeError):
                    indicator = 0
                output[key] = what_store
                output[self.__store_indicator] = indicator
            if key == "_next":
                output[key] = None
                if getattr(response, "next") is not None:
                    output[key] = self.store_keys
        return output

    def from_serializable(self, data: Any) -> Response:
        response = Response()
        for key in self.__response_keys:
            setattr(response, key, data[key])
        for key in self.__response_keys_special:
            if key == "raw":
                response.raw = BytesIO(data[key])
            if key == "headers":
                response.headers = CaseInsensitiveDict(data[key])
            if key == "elapsed":
                response.elapsed = datetime.timedelta(seconds=data[key])
            if key == "_content":
                encoding = response.encoding or self.__implicit_encoding
                indicator = data[self.__store_indicator]
                if indicator == 0:
                    what_store = data[key]
                elif indicator == 1:
                    what_store = data[key].encode(encoding)
                elif indicator == 2:
                    what_store = json.dumps(data[key])
                    what_store = what_store.encode(encoding)
                response._content = what_store  # type: ignore
            if key == "_next":
                setattr(response, "_next", data[key])
        return response



# MIT License
#
# Copyright (c) 2018-2019 Red Hat, Inc.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import inspect
import logging
import shlex
import subprocess
from enum import Enum
from pathlib import Path

from requre.exceptions import PersistentStorageException

logger = logging.getLogger(__name__)


class StorageMode(Enum):
    default = 0
    read = 1
    write = 2
    append = 3


def run_command(cmd, error_message=None, cwd=None, fail=True, output=False):
    """
    subprocess wrapper, copied from packit, for higher level handling of executiong commands,
    :param cmd:
    :param error_message:
    :param cwd:
    :param fail:
    :param output:
    :return:
    """
    if not isinstance(cmd, list):
        cmd = shlex.split(cmd)

    logger.debug("cmd = '%s'", " ".join(cmd))

    cwd = cwd or str(Path.cwd())
    error_message = error_message or f"Command {cmd} failed."

    shell = subprocess.run(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        shell=False,
        cwd=cwd,
        universal_newlines=True,
    )

    if not output:
        # output is returned, let the caller process it
        logger.debug("%s", shell.stdout)
    stderr = shell.stderr.strip()
    if stderr:
        logger.error("%s", shell.stderr)

    if shell.returncode != 0:
        logger.error("Command %s failed", shell.args)
        logger.error("%s", error_message)
        if fail:
            raise PersistentStorageException(
                f"Command {shell.args!r} failed: {error_message}"
            )
        success = False
    else:
        success = True

    if not output:
        return success
    return shell.stdout


class Replacement:
    def __init__(self, name, key, parent, one_filter, replacement) -> None:
        self.name = name
        self.key = key
        self.parent = parent
        self.filter = one_filter
        self.replacement = replacement


def get_module_of_previous_context():
    current_ctx = inspect.currentframe()
    while True:
        current_ctx = current_ctx.f_back
        frameinfo_args = (current_ctx,) + inspect.getframeinfo(current_ctx, 1)
        frameinfo = inspect.FrameInfo(*frameinfo_args)
        module = inspect.getmodule(frameinfo[0])
        if module and not module.__name__.startswith("requre"):
            return module



import tempfile

from requre.helpers.simple_object import Simple
from requre.storage import PersistentObjectStorage
from requre.utils import run_command, StorageMode
from tests.testbase import BaseClass


class StoreFunctionOutput(BaseClass):
    @staticmethod
    @Simple.decorator_plain
    def run_command_wrapper(cmd, error_message=None, cwd=None, fail=True, output=False):
        return run_command(
            cmd=cmd, error_message=error_message, cwd=cwd, fail=fail, output=output
        )

    def test_run_command_true(self):
        """
        Test if session recording is able to store and return output
        from command via decorating run_command
        """
        output = self.run_command_wrapper(cmd=["true"])
        self.assertTrue(output)
        PersistentObjectStorage().dump()
        PersistentObjectStorage().mode = StorageMode.read
        before = str(PersistentObjectStorage().storage_object)
        output = self.run_command_wrapper(cmd=["true"])
        after = str(PersistentObjectStorage().storage_object)
        self.assertTrue(output)
        self.assertIn("True", before)
        self.assertNotIn("True", after)
        self.assertGreater(len(before), len(after))

    def test_run_command_output(self):
        """
         check if wrapper returns proper string values in calls
        """
        self.file_name = tempfile.mktemp()
        with open(self.file_name, "w") as fd:
            fd.write("ahoj\n")
        output = self.run_command_wrapper(cmd=["cat", self.file_name], output=True)
        self.assertIn("ahoj", output)
        PersistentObjectStorage().dump()
        PersistentObjectStorage().mode = StorageMode.read
        with open(self.file_name, "a") as fd:
            fd.write("cao\n")
        output = self.run_command_wrapper(cmd=["cat", self.file_name], output=True)
        self.assertIn("ahoj", output)
        self.assertNotIn("cao", output)
        PersistentObjectStorage().mode = StorageMode.write
        output = self.run_command_wrapper(cmd=["cat", self.file_name], output=True)
        self.assertIn("cao", output)
        PersistentObjectStorage().dump()
        PersistentObjectStorage().mode = StorageMode.read
        output = self.run_command_wrapper(cmd=["cat", self.file_name], output=True)
        self.assertIn("cao", output)



# MIT License
#
# Copyright (c) 2018-2019 Red Hat, Inc.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import builtins
import functools
import re
from enum import Enum
from importlib import reload
from types import ModuleType
from typing import Optional, Dict, Any, Callable, List, Tuple, Union

from requre.utils import Replacement, get_module_of_previous_context


class ReplaceType(Enum):
    """
    Types for import system what are used in replacement list, to know ho to replace it:
    DECORATOR: decorate original function
    REPLACE: replace object by another one
    REPLACE_MODULE: replace whole module by another implementation
    """

    DECORATOR = 1
    REPLACE = 2
    REPLACE_MODULE = 3


def upgrade_import_system(
    filters=None, debug_file: Optional[str] = None
) -> "UpgradeImportSystem":
    """
    High level upgrade import function.

    :param filters: list of filters, for examples see: tests/test_import_system.py
    :param debug_file: file where to store debug information about replacements
    """
    filters = filters or []
    return UpgradeImportSystem(debug_file=debug_file).upgrade(*filters)


def decorate(
    where: str,
    what: str,
    decorator: Callable,
    who_name: Union[str, List[str]] = None,
    debug_file: Optional[str] = None,
) -> "UpgradeImportSystem":
    """
    Decorate the function when importing.

    :param where: which module(s) is(/are) affected
    :param what: what will be decorated
    :param decorator: decorator that will be used
    :param who_name: where is the import change applied
    :param debug_file: file where to store debug information about replacements
    :return: UpgradeImportSystem
    """
    upgraded_import_system = UpgradeImportSystem(debug_file=debug_file)
    upgraded_import_system.decorate(
        where=where, what=what, decorator=decorator, who_name=who_name
    )
    return upgraded_import_system


def replace_module(
    where: str,
    replacement: Any,
    who_name: Union[str, List[str]] = None,
    debug_file: Optional[str] = None,
) -> "UpgradeImportSystem":
    """
    Replace the module when importing.

    :param where: which module(s) is(/are) affected
    :param replacement: what will be used instead
    :param who_name: where is the import change applied
    :param debug_file: file where to store debug information about replacements
    :return: UpgradeImportSystem
    """
    upgraded_import_system = UpgradeImportSystem(debug_file=debug_file)
    upgraded_import_system.replace_module(
        where=where, replacement=replacement, who_name=who_name
    )
    return upgraded_import_system


def replace(
    where: str,
    what: str,
    replacement: Any,
    who_name: Union[str, List[str]] = None,
    debug_file: Optional[str] = None,
) -> "UpgradeImportSystem":
    """
    Replace the module when importing.

    :param where: which module(s) is(/are) affected
    :param what: what will be replaced
    :param replacement: what will be used instead
    :param who_name: where is the import change applied
    :param debug_file: file where to store debug information about replacements
    :return: UpgradeImportSystem
    """
    upgraded_import_system = UpgradeImportSystem(debug_file=debug_file)
    upgraded_import_system.replace(
        where=where, what=what, who_name=who_name, replacement=replacement
    )
    return upgraded_import_system


def log_imports(
    what: str, who_name: Union[str, List[str]] = None, debug_file: Optional[str] = None
) -> "UpgradeImportSystem":
    """
    Log the imports.

    :param what: which module(s) is(/are) affected
    :param who_name: where is the import logged
    :param debug_file: file where to store debug information about replacements
    :return: UpgradeImportSystem
    """
    upgraded_import_system = UpgradeImportSystem(debug_file=debug_file)
    upgraded_import_system.log_imports(what=what, who_name=who_name)
    return upgraded_import_system


class UpgradeImportSystem:
    def __init__(self, debug_file: Optional[str] = None) -> None:
        self.filters: List[Tuple] = []
        self.debug_file = debug_file
        self._original_import = builtins.__import__
        self.replace_dict: Dict[str, Dict[str, Replacement]] = {}

    def __enter__(self) -> "UpgradeImportSystem":
        self.replace_dict.clear()
        self.upgrade()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.revert()

    def revert(self) -> "UpgradeImportSystem":
        """
        Revert the changes to the import system.

        :return: self (chaining is supported)
        """
        builtins.__import__ = self._original_import
        for key_with_replacements in self.replace_dict.values():
            for replacement in key_with_replacements.values():
                if isinstance(replacement.parent, ModuleType):
                    reload(replacement.parent)
        self.replace_dict.clear()
        return self

    def decorate(
        self,
        where: str,
        what: str,
        decorator: Callable,
        who_name: Union[str, List[str]] = None,
    ) -> "UpgradeImportSystem":
        """
        Decorate the function when importing.

        :param where: which module(s) is(/are) affected
        :param what: what will be decorated
        :param decorator: decorator that will be used
        :param who_name: where is the import change applied
        :return: self (chaining is supported)
        """
        if not who_name:
            self.upgrade((where, {}, {what: [ReplaceType.DECORATOR, decorator]}))
            return self

        who_name = who_name if isinstance(who_name, list) else [who_name]
        for who in who_name:
            self.upgrade(
                (where, {"who_name": who}, {what: [ReplaceType.DECORATOR, decorator]})
            )
        return self

    def replace_module(
        self, where: str, replacement: Any, who_name: Union[str, List[str]] = None
    ) -> "UpgradeImportSystem":
        """
        Replace the module when importing.

        :param where: which module(s) is(/are) affected
        :param replacement: what will be used instead
        :param who_name: where is the import change applied
        :return: self (chaining is supported)
        """
        if not who_name:
            self.upgrade((where, {}, {"": [ReplaceType.REPLACE_MODULE, replacement]}))
            return self

        who_name = who_name if isinstance(who_name, list) else [who_name]
        for who in who_name:
            self.upgrade(
                (
                    where,
                    {"who_name": who},
                    {"": [ReplaceType.REPLACE_MODULE, replacement]},
                )
            )
        return self

    def replace(
        self,
        where: str,
        what: str,
        replacement: Any,
        who_name: Union[str, List[str]] = None,
    ) -> "UpgradeImportSystem":
        """
        Replace the module when importing.

        :param where: which module(s) is(/are) affected
        :param what: what will be replaced
        :param replacement: what will be used instead
        :param who_name: where is the import change applied
        :return: self (chaining is supported)
        """

        if not who_name:
            self.upgrade((where, {}, {what: [ReplaceType.REPLACE, replacement]}))
            return self

        who_name = who_name if isinstance(who_name, list) else [who_name]
        for who in who_name:
            self.upgrade(
                (where, {"who_name": who}, {what: [ReplaceType.REPLACE, replacement]})
            )
        return self

    def log_imports(
        self, what: str, who_name: Union[str, List[str]] = None
    ) -> "UpgradeImportSystem":
        """
        Log the imports.

        :param what: which module(s) is(/are) affected
        :param who_name: where is the import logged
        :return: self (chaining is supported)
        """
        if not who_name:
            self.upgrade((what, {}))
            return self

        who_name = who_name if isinstance(who_name, list) else [who_name]
        for who in who_name:
            self.upgrade((what, {"who_name": who}))
        return self

    def upgrade(self, *filters):
        """
        Apply the upgrade to import system in internal-tuple format.

        :param filters: for examples see: tests/test_import_system.py
        :return: self (chaining is supported)
        """

        for filter in filters:
            self.filters.append(filter)
            self._upgrade_filter(builtins.__import__, filter_item=filter)
        return self

    def _upgrade_filter(self, func, filter_item):
        @functools.wraps(func)
        def new_import(*args, **kwargs):

            out = func(*args, **kwargs)
            name = list(args)[0]

            one_filter = filter_item[0]
            additional_filters = filter_item[1]
            if re.search(one_filter, name):
                mod = get_module_of_previous_context()
                fromlist = ()
                if len(args) > 3:
                    fromlist = list(args)[3]
                module_name = getattr(mod, "__name__", "")
                module_file = getattr(mod, "__file__", "")
                item = {
                    "module_object": out,
                    "who": mod,
                    "who_name": module_name,
                    "who_filename": module_file,
                    "fromlist": fromlist,
                }

                if all([re.search(v, item[k]) for k, v in additional_filters.items()]):
                    text = list()
                    text.append(
                        f"{module_name} ({module_file})-> {name} ({fromlist})\n"
                    )
                    if len(filter_item) > 2:
                        for key, replacement in filter_item[2].items():
                            replace_type = replacement[0]
                            replace_object = replacement[1]
                            original_obj = out
                            parent_obj = out
                            # avoid multiple replacing, just in case of module,
                            # because python import system has check
                            # so in case of module it has to be replaced everytime.
                            if (
                                key in self.replace_dict.get(name, {})
                                and replace_type is not ReplaceType.REPLACE_MODULE
                            ):
                                text.append(
                                    f"\t{key} in module {name} already replaced: "
                                    f"{one_filter} -> {key}  by {replacement}\n"
                                )
                            else:
                                # traverse into
                                if len(key) > 0:
                                    for key_item in key.split("."):
                                        parent_obj = original_obj
                                        original_obj = getattr(original_obj, key_item)

                                self.replace_dict.setdefault(name, {})
                                self.replace_dict[name][key] = Replacement(
                                    name=name,
                                    key=key,
                                    one_filter=one_filter,
                                    replacement=replacement,
                                    parent=parent_obj,
                                )

                                if replace_type == ReplaceType.REPLACE:
                                    setattr(
                                        parent_obj,
                                        original_obj.__name__,
                                        replace_object,
                                    )
                                    text.append(
                                        f"\treplacing {key} "
                                        f"by function {replace_object.__name__}\n"
                                    )
                                elif replace_type == ReplaceType.DECORATOR:
                                    setattr(
                                        parent_obj,
                                        original_obj.__name__,
                                        replace_object(original_obj),
                                    )
                                    text.append(
                                        f"\tdecorate {key}  by {replace_object.__name__}\n"
                                    )
                                elif replace_type == ReplaceType.REPLACE_MODULE:
                                    out = replace_object
                                    text.append(
                                        f"\treplace module {name} in {module_name} "
                                        f"by {replace_object.__name__}\n"
                                    )
                    if self.debug_file:
                        with open(self.debug_file, "a") as fd:
                            fd.write("".join(text))
            return out

        builtins.__import__ = new_import



# MIT License
#
# Copyright (c) 2018-2019 Red Hat, Inc.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Any, Optional, Dict
from requre.objects import ObjectStorage


class Simple(ObjectStorage):
    """
    Use this object, when your output is directly YAML serializable, basic objects e.g.
       string, number, boolean, list, dict of these basic objects
    """

    def from_serializable(self, data: Any) -> Any:
        return data

    def to_serializable(self, obj: Any) -> Any:
        return obj


class Void(Simple):
    """
    Use this object, when don't want to store output data of function.
    Could be used for debugging purposes, to see caller arguments
    or output is not important or unable to serialize anyhow.
    """

    def write(self, obj: Any, metadata: Optional[Dict] = None) -> Any:
        """
        Write the object representation to storage
        Internally it will use self.to_serializable()
        method to get serializable object representation

        :param obj: some object
        :param metadata: store metedata to object
        :return: same obj
        """
        self.persistent_storage.store(
            self.store_keys,
            f">>>>> Requre output supressed by using {self.__class__.__name__}",
            metadata=metadata,
        )
        return obj



from requre.import_system import upgrade_import_system
from requre.helpers.simple_object import Simple

FILTERS = upgrade_import_system().decorate(
    where="^time$", what="sleep", who_name=[], decorator=Simple.decorator_plain
)



import builtins
import os
import unittest

from requre.import_system import UpgradeImportSystem, ReplaceType

SELECTOR = os.path.basename(__file__).rsplit(".", 1)[0]


class ContextManager(unittest.TestCase):
    def test_no_filter(self):
        old_imports = builtins.__import__
        with UpgradeImportSystem().upgrade():
            import tempfile

            self.assertNotIn("decorated", tempfile.mktemp())
            self.assertIn("/tmp", tempfile.mktemp())
        assert builtins.__import__ == old_imports

    def test_filter(self):
        old_imports = builtins.__import__
        with UpgradeImportSystem().upgrade(
            (
                "^tempfile$",
                {"who_name": SELECTOR},
                {
                    "mktemp": [
                        ReplaceType.DECORATOR,
                        lambda x: lambda: f"decorated {x()}",
                    ]
                },
            )
        ):
            assert old_imports != builtins.__import__
            import tempfile

            self.assertIn("decorated", tempfile.mktemp())
            self.assertIn("/tmp", tempfile.mktemp())
        assert old_imports == builtins.__import__



import os
import shutil
from requre import RequreTestCase


class CheckBaseTestClass(RequreTestCase):
    def tearDown(self):
        super().tearDown()
        data_file_path = self.get_datafile_filename()
        self.assertTrue(os.path.exists(data_file_path))
        self.assertIn(self.id(), data_file_path)
        self.assertIn("test_data", data_file_path)
        shutil.rmtree(os.path.dirname(os.path.dirname(self.get_datafile_filename())))

    def test(self):
        pass



import warnings
from .simple_object import Simple

warnings.warn(
    "DEPRECATED: Please use Simple.decorator_plain from requre.helpers.simple_object"
)
store_function_output = Simple.decorator_plain



import os
import unittest
import tempfile
import shutil
import time
from requre.utils import run_command
from requre.exceptions import PersistentStorageException
from requre.constants import (
    ENV_REPLACEMENT_FILE,
    ENV_STORAGE_FILE,
    ENV_APPLY_LATENCY,
    ENV_REPLACEMENT_NAME,
)


CMD_RELATIVE = f"""python3 {os.path.join(os.path.dirname(os.path.dirname(__file__)),
                                "requre", "requre_patch.py")}"""
CMD_TOOL = "requre-patch"

DATA_DIR = os.path.join(os.path.dirname(__file__), "data")


def is_sudo_possible():
    random_file = "/usr/bin/test_superuser"
    try:
        run_command(f"sudo -n touch {random_file}")
    except (PersistentStorageException, FileNotFoundError):
        return False
    run_command(f"sudo rm {random_file}")
    return True


def is_requre_installed():
    try:
        run_command(f"{CMD_TOOL} --help")
    except (PersistentStorageException, FileNotFoundError):
        return False
    return True


@unittest.skipUnless(
    is_requre_installed(), "not possible to run without installed requre"
)
class InstalledCommand(unittest.TestCase):
    def setUp(self) -> None:
        run_command(cmd=f"{CMD_RELATIVE} clean", fail=False)

    def tearDown(self) -> None:
        run_command(cmd=f"{CMD_RELATIVE} clean", fail=False)

    def test_not_installed(self):
        self.assertRaises(
            PersistentStorageException, run_command, cmd=f"{CMD_TOOL} verify"
        )

    def test_install(self):
        self.assertIn(
            "Applying import patch",
            run_command(cmd=f"{CMD_TOOL} apply", fail=False, output=True),
        )

    def test_verify_user(self):
        self.test_install()
        output = run_command(cmd=f"{CMD_RELATIVE} verify", output=True)
        self.assertIn("Python patched", output)
        self.assertIn(os.path.expanduser("~/.local/lib/python"), output)

    def test_verify_system(self):
        self.test_install()
        self.assertRaises(
            PersistentStorageException,
            run_command,
            cmd=f"{CMD_RELATIVE} --system verify",
        )


class NoApplied(unittest.TestCase):
    def setUp(self) -> None:
        run_command(cmd=f"{CMD_RELATIVE} clean", fail=False)

    def test_help(self):
        run_command(cmd=f"{CMD_RELATIVE}")

    def test_bad_command(self):
        self.assertRaises(
            PersistentStorageException, run_command, cmd=f"{CMD_RELATIVE} badcommand"
        )

    def test_verify(self):
        self.assertRaises(
            PersistentStorageException, run_command, cmd=f"{CMD_RELATIVE} verify"
        )


class Apply(unittest.TestCase):
    def setUp(self) -> None:
        run_command(cmd=f"{CMD_RELATIVE} clean", fail=False)

    def tearDown(self) -> None:
        run_command(cmd=f"{CMD_RELATIVE} clean", fail=False)

    def test_apply(self):
        run_command(cmd=f"{CMD_RELATIVE} apply")

    def test_verify(self):
        self.test_apply()
        run_command(cmd=f"{CMD_RELATIVE} verify")

    def test_clean(self):
        self.test_verify()
        run_command(cmd=f"{CMD_RELATIVE} clean")


class NormalUser(unittest.TestCase):
    test_command = f"python3 {DATA_DIR}/e2e_test.py"

    def setUp(self) -> None:
        run_command(cmd=f"{CMD_RELATIVE} clean", fail=False)
        run_command(cmd=f"{CMD_RELATIVE} apply")
        self.storage_file = os.path.join(tempfile.mktemp(), "storage_e2e.yaml")

    def tearDown(self) -> None:
        run_command(cmd=f"{CMD_RELATIVE} clean", fail=False)
        if os.path.exists(self.storage_file):
            os.remove(self.storage_file)
        tmpdir = os.path.join("/tmp", os.path.basename(self.storage_file))
        if os.path.exists(tmpdir):
            shutil.rmtree(tmpdir)

    def testWithoutSupport(self):
        output = run_command(cmd=self.test_command, output=True)
        self.assertIn("/tmp/tmp", output)
        self.assertNotIn("static_tmp_1", output)
        self.assertNotIn("storage_e2e", output)

    def testSupport(self):
        envs = (
            f"{ENV_STORAGE_FILE}={self.storage_file}"
            f" {ENV_REPLACEMENT_FILE}={DATA_DIR}/e2e_test_replacements.py"
        )
        cmd = f"""bash -c '{envs} {self.test_command}'"""
        output = run_command(cmd=cmd, output=True)
        self.assertIn("static_tmp_1", output)
        self.assertIn("storage_e2e", output)


@unittest.skipUnless(
    is_sudo_possible(), "not possible to run, unless you are root or have sudo"
)
class SuperUser(NormalUser):
    test_command = f"python3 {DATA_DIR}/e2e_test.py"

    def setUp(self) -> None:
        run_command(cmd=f"sudo {CMD_RELATIVE} --system clean", fail=False)
        run_command(cmd=f"sudo {CMD_RELATIVE} --system apply")
        self.storage_file = os.path.join(tempfile.mktemp(), "storage_e2e.yaml")

    def tearDown(self) -> None:
        run_command(cmd=f"sudo {CMD_RELATIVE} --system clean", fail=False)
        if os.path.exists(self.storage_file):
            os.remove(self.storage_file)
        tmpdir = os.path.join("/tmp", os.path.basename(self.storage_file))
        if os.path.exists(tmpdir):
            shutil.rmtree(tmpdir)

    def test_installation_path(self):
        output = run_command(cmd=f"{CMD_RELATIVE} --system verify", output=True)
        self.assertIn("Python patched", output)
        self.assertIn("/usr/lib/python", output)


class Latency(unittest.TestCase):
    test_command = f"python3 {DATA_DIR}/e2e_latency_test.py"

    def setUp(self) -> None:
        run_command(cmd=f"{CMD_RELATIVE} clean", fail=False)
        run_command(cmd=f"{CMD_RELATIVE} apply")
        self.storage_file = tempfile.mktemp()

    def tearDown(self) -> None:
        run_command(cmd=f"{CMD_RELATIVE} clean", fail=False)
        os.remove(self.storage_file)

    def check(self, replacements, apply_latency, latency, delta):
        """
        Check if it is really waiting for function call, when latency is enabled
        """
        envs = (
            f"{ENV_STORAGE_FILE}={self.storage_file} "
            f"{ENV_REPLACEMENT_FILE}={DATA_DIR}/{replacements} "
            f"{ENV_APPLY_LATENCY}={apply_latency}"
        )
        cmd = f"""bash -c '{envs} {self.test_command}'"""
        self.assertFalse(os.path.exists(self.storage_file))
        before = time.time()
        run_command(cmd=cmd, output=True)
        after = time.time()
        self.assertAlmostEqual(2, after - before, delta=delta)

        self.assertTrue(os.path.exists(self.storage_file))
        before = time.time()
        run_command(cmd=cmd, output=True)
        after = time.time()
        self.assertAlmostEqual(latency, after - before, delta=delta)

    def test_not_enabled(self):
        self.check("e2e_latency_replacements.py", "", 0, 1)

    def test_enabled_plain(self):
        self.check("e2e_latency_replacements.py", "YES", 2, 1)

    def test_enabled_object_model(self):
        self.check("e2e_latency_replacements_object.py", "YES", 2, 1)


class ReplacementVariable(unittest.TestCase):
    test_command = f"python3 {DATA_DIR}/e2e_test.py"

    def setUp(self) -> None:
        run_command(cmd=f"{CMD_RELATIVE} clean", fail=False)
        run_command(cmd=f"{CMD_RELATIVE} apply")
        self.storage_file = os.path.join(tempfile.mktemp(), "storage_e2e.yaml")

    def tearDown(self) -> None:
        run_command(cmd=f"{CMD_RELATIVE} clean", fail=False)
        if os.path.exists(self.storage_file):
            os.remove(self.storage_file)
        tmpdir = os.path.join("/tmp", os.path.basename(self.storage_file))
        if os.path.exists(tmpdir):
            shutil.rmtree(tmpdir)

    def testSupport(self):
        envs = (
            f"{ENV_STORAGE_FILE}={self.storage_file}"
            f" {ENV_REPLACEMENT_FILE}={DATA_DIR}/e2e_test_replacements_special.py"
            f" {ENV_REPLACEMENT_NAME}=special"
        )
        cmd = f"""bash -c '{envs} {self.test_command}'"""
        output = run_command(cmd=cmd, output=True)
        self.assertIn("static_tmp_1", output)
        self.assertIn("storage_e2e", output)

    def testException(self):
        envs = (
            f"{ENV_STORAGE_FILE}={self.storage_file}"
            f" {ENV_REPLACEMENT_FILE}={DATA_DIR}/e2e_test_replacements_special.py"
            f" {ENV_REPLACEMENT_NAME}=nonsense"
        )
        cmd = f"""bash -c '{envs} {self.test_command}'"""
        self.assertRaises(PersistentStorageException, run_command, cmd=cmd, output=True)



# MIT License
#
# Copyright (c) 2018-2019 Red Hat, Inc.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import functools
import logging
import os
import tarfile
from io import BytesIO
from typing import Callable, Any, Dict

from requre.exceptions import PersistentStorageException
from requre.helpers.simple_object import Simple
from requre.storage import PersistentObjectStorage, DataMiner

logger = logging.getLogger(__name__)


class StoreFiles:
    dir_suffix = "file_storage"
    tar_compression = "xz"
    basic_ps_keys = ["X", "file", "tar"]

    @classmethod
    def _test_identifier(cls):
        return os.path.basename(PersistentObjectStorage().storage_file)

    @classmethod
    def _copy_logic(
        cls, pers_storage: PersistentObjectStorage, pathname: str, keys: list
    ) -> None:
        """
        Internal function. Copy files to or back from persisten storage
        It will  create tar archive with tar_compression and stores it to Persistent Storage
        """
        logger.debug(f"Copy files {pathname} -> {keys}")
        logger.debug(f"Persistent Storage mode: {pers_storage.mode}")
        original_cwd = os.getcwd()
        if pers_storage.do_store(keys=cls.basic_ps_keys + keys):
            try:
                artifact_name = os.path.basename(pathname)
                artifact_path = os.path.dirname(pathname)
                os.chdir(artifact_path)
                with BytesIO() as fileobj:
                    with tarfile.open(
                        mode=f"x:{cls.tar_compression}", fileobj=fileobj
                    ) as tar_store:
                        tar_store.add(name=artifact_name)
                    fileobj.seek(0)
                    metadata = {DataMiner().LATENCY_KEY: 0}
                    pers_storage.store(
                        keys=cls.basic_ps_keys + keys,
                        values=fileobj.read(),
                        metadata=metadata,
                    )
            finally:
                os.chdir(original_cwd)
        else:
            value = pers_storage[cls.basic_ps_keys + keys]
            with BytesIO(value) as fileobj:
                with tarfile.open(
                    mode=f"r:{cls.tar_compression}", fileobj=fileobj
                ) as tar_store:
                    tarinfo_1st_member = tar_store.getmembers()[0]
                    if tarinfo_1st_member.isfile():
                        with open(pathname, mode="wb") as output_file:
                            output_file.write(
                                tar_store.extractfile(tarinfo_1st_member).read()
                            )
                    else:
                        for tar_item in tar_store.getmembers():
                            # we have to modify path of files to remove topdir
                            if len(tar_item.name.split(os.path.sep, 1)) > 1:
                                tar_item.name = tar_item.name.split(os.path.sep, 1)[1]
                            else:
                                tar_item.name = "."
                            try:
                                tar_store.extract(tar_item, path=pathname)
                            except IOError:
                                # rewrite readonly files if necessary
                                os.remove(os.path.join(pathname, tar_item.name))
                                tar_store.extract(tar_item, path=pathname)

    @classmethod
    def return_value(cls, func: Callable) -> Any:
        """
        Decorator what will store return value of function/method as file and will store content

        """

        @functools.wraps(func)
        def store_files_int(*args, **kwargs):
            output = Simple.decorator_plain(func)(*args, **kwargs)
            cls._copy_logic(
                PersistentObjectStorage(),
                pathname=output,
                keys=[cls.__name__, cls._test_identifier()],
            )
            return output

        return store_files_int

    @classmethod
    def guess_args(cls, func: Callable) -> Any:
        """
        Decorator what try to guess, which arg is file or directory and store its content
        """

        @functools.wraps(func)
        def store_files_int(*args, **kwargs):
            def int_dec_fn(pathname_arg, keys_arg):
                if not isinstance(pathname_arg, str):
                    return
                if PersistentObjectStorage().do_store(
                    keys=StoreFiles.basic_ps_keys + keys_arg
                ):
                    if os.path.exists(pathname_arg):
                        cls._copy_logic(
                            PersistentObjectStorage(),
                            pathname=pathname_arg,
                            keys=keys_arg,
                        )
                else:
                    try:
                        cls._copy_logic(
                            PersistentObjectStorage(),
                            pathname=pathname_arg,
                            keys=keys_arg,
                        )
                    except PersistentStorageException:
                        pass

            class_test_id_list = [cls.__name__, cls._test_identifier()]
            output = Simple.decorator_plain(func)(*args, **kwargs)
            for position in range(len(args)):
                int_dec_fn(args[position], class_test_id_list + [position])
            for k, v in kwargs.items():
                int_dec_fn(v, class_test_id_list + [k])
            return output

        return store_files_int

    @classmethod
    def arg_references(cls, files_params: Dict) -> Any:
        """
        Decorator what will store files or directory based on arguments,
        you have to pass name and position of arg via dict
        (be careful about counting self or cls parameteres for methods)
        eg. files_params = {"target_dir": 2}
        """

        def store_files_int(func):
            @functools.wraps(func)
            def store_files_int_int(*args, **kwargs):
                class_test_id_list = [cls.__name__, cls._test_identifier()]
                output = Simple.decorator_plain(func)(*args, **kwargs)
                for key, position in files_params.items():
                    if key in kwargs:
                        param = kwargs[key]
                    else:
                        param = args[position]
                    cls._copy_logic(
                        PersistentObjectStorage(),
                        pathname=param,
                        keys=class_test_id_list + [key],
                    )
                return output

            return store_files_int_int

        return store_files_int

    @classmethod
    def explicit_reference(cls, file_param: str, dest_key: str = "default") -> Any:
        """
        Method to store explicitly path file_param to persistent storage
        """
        class_test_id_list = [cls.__name__, cls._test_identifier()]
        cls._copy_logic(
            PersistentObjectStorage(),
            pathname=file_param,
            keys=class_test_id_list + [dest_key],
        )



from requre.helpers.requests_response import RequestResponseHandling
from requre.import_system import ReplaceType

session_send = {
    "Session.send": [
        ReplaceType.DECORATOR,
        RequestResponseHandling.decorator(item_list=[]),
    ]
}

MODULE_LIST = [
    ("^requests$", {"who_name": "ogr"}),
    ("^requests$", {"who_name": "gitlab"}),
    ("^requests$", {"who_name": "github"}),
    ("^requests$", {"who_name": "ogr.services.pagure"}, session_send),
    ("^requests$", {"who_name": "gitlab"}, session_send),
    ("^requests$", {"who_name": "github.MainClass"}, session_send),
    ("^requests$", {"who_name": "github.Requester"}, session_send),
    ("^requests$", {"who_name": "ogr.services.github_tweak"}, session_send),
]



# MIT License
#
# Copyright (c) 2018-2019 Red Hat, Inc.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from typing import Any, Callable, List

from git.remote import FetchInfo
from git.util import IterableList
from requre.storage import PersistentObjectStorage
from requre.objects import ObjectStorage
from requre.helpers.files import StoreFiles


class FetchInfoStorageList(ObjectStorage):
    # TODO: improve handling of "ref" item (need deep inspection of git objects and consequences)
    # it is not mandatory for current packit operations
    __ignored = ["ref"]
    __response_keys_special: List[str] = []
    __response_keys = list(
        set(FetchInfo.__slots__) - set(__ignored) - set(__response_keys_special)
    )

    object_type = IterableList

    def to_serializable(self, obj: Any) -> Any:
        output = list()
        for item in obj:
            tmp = dict()
            for key in self.__response_keys:
                tmp[key] = getattr(item, key)
            output.append(tmp)
        return output

    def from_serializable(self, data: Any) -> Any:
        out = self.object_type("name")
        for item in data:
            tmp = FetchInfo(None, None, None, None)
            for key in self.__response_keys:
                setattr(tmp, key, item[key])
            out.append(tmp)
        return out


class RemoteFetch(FetchInfoStorageList):
    """
    Use this class for git.remote.Remote.fetch recording
    """

    @classmethod
    def execute(cls, keys: list, func: Callable, *args, **kwargs) -> Any:
        """
        Class method to store or read object from persistent storage
        with explicit set of *args, **kwargs parameters to use as keys
        :param keys: list of keys used as parameter
        :param func: original function
        :param args: parameters of original function
        :param kwargs: parameters of original function
        :return: output of called func
        """

        output = super().execute(keys, func, *args, **kwargs)
        git_object = args[0]
        git_dir = git_object.repo.git_dir
        StoreFiles.explicit_reference(git_dir)
        if not PersistentObjectStorage().do_store(keys):
            # mimic the code in git for read mode for Remote.fetch
            # https://github.com/gitpython-developers/GitPython/blob/master/git/remote.py
            if hasattr(git_object.repo.odb, "update_cache"):
                git_object.repo.odb.update_cache()
        return output



# MIT License
#
# Copyright (c) 2018-2019 Red Hat, Inc.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import inspect
import logging
import os
import time
from _collections_abc import Hashable
from enum import Enum
from typing import Dict, List, Any, Optional, Callable

import yaml

from .constants import (
    VERSION_REQURE_FILE,
    ENV_STORAGE_FILE,
    KEY_MINIMAL_MATCH,
    METATADA_KEY,
    ENV_REQURE_STORAGE_MODE,
)
from .exceptions import (
    ItemNotInStorage,
    StorageNoResponseLeft,
    PersistentStorageException,
)
from .singleton import SingletonMeta
from .utils import StorageMode

# use this sleep to avoid decorating original time function used internally
original_sleep = time.sleep
original_time = time.time

logger = logging.getLogger(__name__)


class StorageKeysInspect:
    @staticmethod
    def get_base_keys(func: Callable) -> List[Any]:
        raise NotImplementedError("Use child classes")


class StorageKeysInspectFull(StorageKeysInspect):
    @staticmethod
    def get_base_keys(func: Callable) -> List[Any]:
        output: List[str] = list()
        # callers module list, to be able to separate requests for various services in one file
        caller_list: List[str] = list()
        for currnetframe in inspect.stack():
            if not inspect.getmodule(currnetframe[0]):
                continue
            module_name = inspect.getmodule(currnetframe[0]).__name__
            if module_name.startswith("_"):
                break
            # avoid to store requre.storage to module stack
            # backward compatibility issue
            if module_name.startswith("requre.storage"):
                continue
            if len(caller_list) and caller_list[-1] == module_name:
                continue
            else:
                caller_list.append(module_name)
        output += caller_list[::-1]
        # module name where function is
        output.append(inspect.getmodule(func).__name__)
        # name of function what were used
        output.append(func.__name__)
        return output


class StorageKeysInspectSimple(StorageKeysInspect):
    @staticmethod
    def get_base_keys(func: Callable) -> List[Any]:
        return [inspect.getmodule(func).__name__, func.__name__]


class StorageKeysInspectUnique(StorageKeysInspect):
    @staticmethod
    def get_base_keys(func: Callable) -> List[Any]:
        return StorageKeysInspectUnique._get_unique_keys(
            StorageKeysInspectFull.get_base_keys(func)
        )

    @staticmethod
    def _get_unique_keys(keys):
        output: List = []
        for item in reversed(keys):
            if item not in output:
                output.append(item)
        return list(reversed(output))


StorageKeysInspectDefault = StorageKeysInspectFull


class DataStructure:
    """
    Object model for storing data to persistent storage
    """

    OUTPUT_KEY = "output"
    METADATA_KEY = "metadata"

    def __init__(self, output: Any):
        self.output = output
        self._metadata: dict = dict()

    @property
    def metadata(self):
        """
        Medatata for item in storage
        """
        return self._metadata

    @metadata.setter
    def metadata(self, values: dict):
        for k, v in values.items():
            self._metadata[k] = v

    def dump(self):
        """
        Dump object to serializable format

        :return dict
        """
        return {self.METADATA_KEY: self.metadata, self.OUTPUT_KEY: self.output}

    @classmethod
    def create_from_value(cls, dict_repr: dict):
        """
        Create Object representation from dict

        :return DataStructure
        """
        data = cls(dict_repr[cls.OUTPUT_KEY])
        data.metadata = dict_repr[cls.METADATA_KEY]
        return data

    @classmethod
    def create_from_dict(cls, dict_repr: dict):
        """
        Create Object representation from dict

        :return DataStructure
        """
        if DataMiner().key not in dict_repr:
            raise ItemNotInStorage(
                f"Key '{DataMiner().key}' not in the response file. "
                f"(Be sure that you generate the response file "
                f"for this variant.)"
            )
        value = dict_repr[DataMiner().key]
        data = cls(value[cls.OUTPUT_KEY])
        data.metadata = value[cls.METADATA_KEY]
        return data

    @classmethod
    def create_from_dict_with_list(cls, dict_repr: dict):
        """
        Create Object representation from dict with list

        :return DataStructure
        """
        if DataMiner().key not in dict_repr:
            raise ItemNotInStorage(
                f"Key '{DataMiner().key}' not in the response file. "
                f"(Be sure that you generate the response file "
                f"for this variant.)"
            )
        value = dict_repr[DataMiner().key]
        return cls.create_from_list(list_repr=value)

    @classmethod
    def create_from_list(cls, list_repr: list):
        if len(list_repr) == 0:
            raise StorageNoResponseLeft(
                "No responses left. Try to regenerate response file "
                f"({PersistentObjectStorage().storage_file})."
            )
        if PersistentObjectStorage().storage_file_version <= 1:
            # Backward compatibility for requre before using DataStructure
            data = DataStructure(list_repr[0])
        else:
            data = DataStructure.create_from_value(list_repr[0])
        del list_repr[0]
        return data


class DataTypes(Enum):
    List = 1
    Value = 2
    Dict = 3
    DictWithList = 4


class DataMiner(metaclass=SingletonMeta):
    """
    Intermediate class used for proper formatting and keeping backward
    compatibility with requre storage version files
    """

    def __init__(self):
        self.current_time = original_time()
        self.data: Optional[DataStructure] = None
        self.data_type: DataTypes = DataTypes.List
        self.use_latency = False
        self.LATENCY_KEY = "latency"
        self.key: str = "all"
        self.key_stategy_cls = StorageKeysInspectDefault
        self.store_arg_debug_metadata = False
        self.METADATA_ARG_DEBUG_KEY = "log_call_function"
        self.METADATA_CALLER_LIST = "module_call_list"
        self.read_key_exact = False

    def get_latency(self, regenerate=True) -> float:
        """
        Returns latency from last store event

        :param regenerate: if False, then it will not be changed, just displayed
        :return: float
        """
        current_time = original_time()
        diff = current_time - self.current_time
        if regenerate:
            self.current_time = current_time
        return diff

    def dump(self, level, key, values, metadata: Dict):
        """
        Store values in proper format to level[key] item

        :param level: parent dict object
        :param key: item in dict
        :param values: what to store (return value of function)
        :param metadata: store metadata to object from upper level
        :return: None
        """
        ds = DataStructure(values)
        if metadata:
            ds.metadata = metadata
        if self.LATENCY_KEY not in ds.metadata:
            ds.metadata = {self.LATENCY_KEY: self.get_latency()}
        self.data = ds
        item = ds.dump()
        if self.data_type == DataTypes.Dict:
            level.setdefault(key, {})
            level[key][self.key] = item
        elif self.data_type == DataTypes.DictWithList:
            level.setdefault(key, {})
            level[key].setdefault(self.key, [])
            level[key][self.key].append(item)
        elif self.data_type in [DataTypes.List, DataTypes.DictWithList]:
            if PersistentObjectStorage().storage_file_version <= 1:
                # Backward compatibility for requre before using DataStructure
                item = values
            level.setdefault(key, [])
            level[key].append(item)
        elif self.data_type == DataTypes.Dict:
            level.setdefault(key, {})
            level[key][self.key] = item
        else:
            level[key] = item

    def load(self, level):
        """
        Get data from storage_object and trasform it to DataStructure object.

        :param level: where data are stored
        :return: output of function
        """
        data = None
        if self.data_type == DataTypes.List:
            data = DataStructure.create_from_list(level)
        elif self.data_type == DataTypes.Value:
            data = DataStructure.create_from_value(level)
        elif self.data_type == DataTypes.Dict:
            data = DataStructure.create_from_dict(level)
        elif self.data_type == DataTypes.DictWithList:
            data = DataStructure.create_from_dict_with_list(level)
        self.data = data
        if self.use_latency:
            original_sleep(data.metadata.get(self.LATENCY_KEY, 0))
        return data.output

    @property
    def metadata(self):
        return self.data.metadata

    @metadata.setter
    def metadata(self, value):
        self.data.metadata = value


class PersistentObjectStorage(metaclass=SingletonMeta):
    """
    Class implements reading/writing simple JSON requests to dict structure
    and return values based on keys.
    It contains methods to reads/stores data to object and load and store them to YAML file

    storage_object: dict with structured data based on keys (eg. simple JSON requests)
    storage_file: file for reading and writing data in storage_object
    """

    internal_object_key = METATADA_KEY
    version_key = "version_storage_file"
    key_inspect_strategy_key = "key_strategy"

    def _set_defaults(self) -> None:
        self.dump_after_store = False
        storage_mode = os.getenv(ENV_REQURE_STORAGE_MODE, "default")
        logger.info(
            f"You overrided storage mode via env var {ENV_REQURE_STORAGE_MODE}={storage_mode}"
        )
        if not hasattr(StorageMode, storage_mode):
            raise PersistentStorageException(
                f"storage mode '{storage_mode}' does not exist, "
                f"use one of {list(StorageMode.__members__.keys())}) "
            )
        self.mode: StorageMode = getattr(StorageMode, storage_mode)
        self.is_flushed = False
        self.storage_object: dict = {}
        self._storage_file: Optional[str] = None

    def __init__(self) -> None:
        # call dump() after store() is called
        self._set_defaults()
        storage_file_from_env = os.getenv(ENV_STORAGE_FILE)
        if storage_file_from_env:
            self.storage_file = storage_file_from_env

    @property
    def metadata(self):
        """
        Placeholder to store some custom configuration, or store custom data,
        eg. version of storage file, or some versions of packages

        :return: dict
        """
        return self.storage_object.get(self.internal_object_key, {})

    @metadata.setter
    def metadata(self, key_dict: dict):
        if self.internal_object_key not in self.storage_object:
            self.storage_object[self.internal_object_key] = {}
        for k, v in key_dict.items():
            self.storage_object[self.internal_object_key][k] = v
        self.is_flushed = False

    @property
    def storage_file_version(self):
        """
        Get version of persistent storage file
        :return: int
        """
        return self.metadata.get(self.version_key, 0)

    def do_store(self, keys):
        if self.mode == StorageMode.read and not os.path.exists(self.storage_file):
            raise PersistentStorageException(
                "Requre can't work in this setup: we are meant to read "
                "recorded responses but the storage file does not exist."
            )
        if self.mode == StorageMode.read:
            return False

        elif self.mode == StorageMode.write:
            return True

        elif self.mode == StorageMode.append:
            if DataMiner().data_type in [DataTypes.DictWithList, DataTypes.List]:
                # append possible
                return True

            if (
                DataMiner().data_type in [DataTypes.Dict, DataTypes.Value]
                and keys not in self
            ):
                # keys present => we can read the values
                return True
            return False

        key_present = keys in self
        logger.warning(
            f"No storage-mode set. Checking presence of the keys: {key_present} "
            f"=> {'read' if key_present else 'store'}"
        )
        return not key_present

    @storage_file_version.setter  # type: ignore
    def storage_file_version(self, value: int):
        self.metadata = {self.version_key: value}

    def _set_storage_metadata_if_not_set(self):
        """
        Set storage file version if not already set to latest version.
        """
        if self.metadata.get(self.version_key) is None:
            self.storage_file_version = VERSION_REQURE_FILE
        if self.metadata.get(self.key_inspect_strategy_key) is None:
            self.metadata = {
                self.key_inspect_strategy_key: DataMiner().key_stategy_cls.__name__
            }
        data_type_name = DataTypes.__name__
        if self.metadata.get(data_type_name) is None:
            self.metadata = {data_type_name: DataMiner().data_type.value}

    @property
    def storage_file(self):
        return self._storage_file

    @storage_file.setter
    def storage_file(self, value):
        # when file is changed set PersistenStorage default values
        if self._storage_file != value:
            self._set_defaults()
            self._storage_file = value

        if not os.path.exists(self._storage_file):
            if self.mode == StorageMode.default:
                self.mode = StorageMode.write

            elif self.mode == StorageMode.read:
                raise PersistentStorageException(
                    "Requre can't work in this setup: we are meant to read"
                    f" recorded responses but the storage file ({self._storage_file}) "
                    "does not exist."
                )
        else:
            if self.mode == StorageMode.default:
                self.mode = StorageMode.read
            self.storage_object = self.load()

    @staticmethod
    def transform_hashable(keys: List) -> List:
        output: List = []
        for item in keys:
            if not item:
                output.append("empty")
            elif not isinstance(item, Hashable):
                output.append(str(item))
            else:
                output.append(item)
        return output

    def store(self, keys: List, values: Any, metadata: Dict) -> None:
        """
        Stores data to dictionary object based on keys values it will create structure
        if structure does not exist

        It implicitly changes type to string if key is not hashable

        :param keys: items what will be used as keys for dictionary
        :param values: It could be whatever type what is used in original object handling
        :return: None
        """
        self._set_storage_metadata_if_not_set()
        current_level = self.storage_object
        hashable_keys = self.transform_hashable(keys)
        for item_num in range(len(hashable_keys)):
            item = hashable_keys[item_num]
            if item_num + 1 < len(hashable_keys):
                if not current_level.get(item):
                    current_level[item] = {}
            else:
                DataMiner().dump(
                    level=current_level, key=item, values=values, metadata=metadata
                )
            current_level = current_level[item]
        self.is_flushed = False

        if self.dump_after_store:
            self.dump()
        logger.debug(f"Storing response to: {self.storage_file}: {hashable_keys}")

    def read(self, keys: List) -> Any:
        """
        Reads data from dictionary object structure based on keys.
        If keys does not exists

        It implicitly changes type to string if key is not hashable

        :param keys: key list for searching in dict
        :return: value assigged to key items
        """
        current_level = self.storage_object
        hashable_keys = self.transform_hashable(keys)
        debug_keys: List[str] = []
        matched_calls: List[str] = []
        list_len = len(hashable_keys)
        for item_num in range(list_len):
            item = hashable_keys[item_num]
            if item not in current_level:
                # it matched last 2 items
                if DataMiner().read_key_exact or (
                    not DataMiner().read_key_exact
                    and item_num + KEY_MINIMAL_MATCH >= list_len
                ):
                    # if not matched, but consider if it is not same key as previous
                    # it is important if simplify used.
                    if matched_calls and item == matched_calls[-1]:
                        debug_keys.append(f"DUPLICATE {item}")
                        continue
                    raise ItemNotInStorage(
                        f"Keys not in storage:{self.storage_file}"
                        f" Matched: {debug_keys},"
                        f" Missing: {hashable_keys[item_num:]}"
                    )
                else:
                    debug_keys.append(f"SKIP {item}")

            else:
                debug_keys.append(item)
                matched_calls.append(item)
                current_level = current_level[item]
        try:
            result = DataMiner().load(level=current_level)
        except StorageNoResponseLeft as e:
            raise StorageNoResponseLeft(f"{e.args} (keys: {debug_keys})")
        logger.debug(f"Reading response from: {self.storage_file}: {debug_keys}")
        return result

    def __contains__(self, item) -> bool:
        current_level = self.storage_object
        hashable_keys = self.transform_hashable(item)
        for item in hashable_keys:
            if item not in current_level:
                return False
            current_level = current_level[item]
        if DataMiner().data_type in [DataTypes.Dict, DataTypes.DictWithList]:
            return DataMiner().key in current_level
        return True

    def __getitem__(self, key):
        return self.read(keys=key)

    def __setitem__(self, key, value):
        self.store(keys=key, values=value, metadata={})

    def __delitem__(self, key):
        current_level = self.storage_object
        hashable_keys = self.transform_hashable(key)
        last_level = None
        for item in hashable_keys:
            if item not in current_level:
                raise ItemNotInStorage(
                    f"Key not found in the persistent storage: {key}"
                )
            last_level = current_level
            current_level = current_level[item]

        if DataMiner().data_type == DataTypes.Dict:
            del current_level[DataMiner().key]
        else:
            del last_level[key[-1]]

    def dump(self) -> None:
        """
        Explicitly stores content of storage_object to storage_file path

        This method is also called when object is deleted and is set write mode to True

        :return: None
        """
        if self.mode in [StorageMode.write, StorageMode.append]:
            self._set_storage_metadata_if_not_set()
            if self.is_flushed:
                return None
            with open(self.storage_file, "w") as yaml_file:
                yaml.dump(self.storage_object, yaml_file, default_flow_style=False)
            self.is_flushed = True

    def load(self) -> Dict:
        """
        Explicitly loads file content of storage_file to storage_object and return as well

        :return: dict
        """
        with open(self.storage_file, "r") as yaml_file:
            output = yaml.safe_load(yaml_file)
        self.storage_object = output
        # set proper storage strategy if stored in file
        if self.metadata.get(self.key_inspect_strategy_key):
            logger.debug(
                f"Used key strategy: {self.metadata.get(self.key_inspect_strategy_key)}"
            )
            DataMiner().key_stategy_cls = globals()[
                self.metadata.get(self.key_inspect_strategy_key)
            ]
        return output


def use_persistent_storage_without_overwriting(cls):
    class ClassWithPersistentStorage(cls):
        persistent_storage: Optional[
            PersistentObjectStorage
        ] = PersistentObjectStorage()

    ClassWithPersistentStorage.__name__ = cls.__name__
    return ClassWithPersistentStorage


class StorageCounter:
    counter = 0
    dir_suffix = "file_storage"
    previous = None

    @classmethod
    def reset_counter_if_changed(cls):
        current = os.path.basename(PersistentObjectStorage().storage_file)
        if cls.previous != current:
            cls.counter = 0
            cls.previous = current

    @classmethod
    def next(cls):
        cls.counter += 1
        return cls.counter

    @staticmethod
    def storage_file():
        return os.path.basename(PersistentObjectStorage().storage_file)

    @staticmethod
    def storage_dir():
        return os.path.dirname(PersistentObjectStorage().storage_file)



import os
import unittest
from copy import deepcopy
from requre.utils import run_command
from requre.postprocessing import DictProcessing
from requre.storage import PersistentObjectStorage, DataTypes, DataMiner
from tests.testbase import BaseClass
from tests.test_e2e_test_patching import is_requre_installed

CMD_TOOL = "requre-patch purge"
DATA_DIR = os.path.join(os.path.dirname(__file__), "data")


class ObjectPostprocessing(unittest.TestCase):
    def setUp(self) -> None:
        self.testDict = {
            "a": {"b": {"c": "ahoj"}},
            "int": 2,
            "str": "anystr",
            "bool": False,
            "list": [1, 2, {"a": "y"}],
            "list_inside": [{"m": {"n": "o"}}],
            "dict": {"a": "x"},
        }
        self.dp = DictProcessing(self.testDict)

    def testMatchNoSelector(self):
        self.assertEqual(1, len(list(self.dp.match([]))))

    def testMatchMultiSelector(self):
        self.assertEqual(3, len(list(self.dp.match(["a"]))))

    def testMatchDeep(self):
        output = list(self.dp.match(["list_inside", "m"]))
        self.assertEqual(1, len(output))
        self.assertEqual({"n": "o"}, output[0])

        output = list(self.dp.match(["n"]))
        self.assertEqual(1, len(output))
        self.assertEqual("o", output[0])

    def testMatchNoMatch(self):
        self.assertEqual(0, len(list(self.dp.match(["non_sense"]))))

    def testReplaceMulti(self):
        tmp_dict = deepcopy(self.testDict)
        DictProcessing.replace(self.testDict, "a", "some")
        self.assertEqual(self.testDict["a"], "some")
        self.assertEqual(self.testDict["dict"]["a"], "some")
        self.assertEqual(self.testDict["list"][2]["a"], "some")
        self.assertNotEqual(tmp_dict, self.testDict)

    def testReplaceNone(self):
        tmp_dict = deepcopy(self.testDict)
        DictProcessing.replace(self.testDict, "nonsense", "some")
        self.assertEqual(tmp_dict, self.testDict)


@unittest.skipUnless(
    is_requre_installed(), "not possible to run without installed requre"
)
class FilePostprocessing(unittest.TestCase):
    storage_file = os.path.join(DATA_DIR, "requre_postprocessing.yaml")
    storage_file_tmp = storage_file + ".tmp"

    def setUp(self) -> None:
        run_command(cmd=f"cp {self.storage_file} {self.storage_file_tmp}")

    def tearDown(self) -> None:
        run_command(cmd=f"rm {self.storage_file_tmp}")

    def testOutput(self):
        replacement = (
            "tests_requre.openshift_integration.test_fedpkg.FedPkg."
            "test_fedpkg_clone.yaml:output:str:random_output"
        )
        run_command(f"{CMD_TOOL} --replaces {replacement} {self.storage_file_tmp}")
        with open(self.storage_file_tmp, mode="r") as opened_file:
            output = opened_file.read()
            self.assertIn("random_output", output)
            self.assertIn("requre.openshift_integration", output)
            self.assertIn("latency: xxxx", output)
            self.assertNotIn("output: yyy", output)

    def testLatency(self):
        replacement = "clone%store_function_output:latency:int:50"
        run_command(f"{CMD_TOOL} --replaces {replacement} {self.storage_file_tmp}")
        with open(self.storage_file_tmp, mode="r") as opened_file:
            output = opened_file.read()
            self.assertNotIn("random_output", output)
            self.assertIn("requre.openshift_integration", output)
            self.assertNotIn("latency: xxxx", output)
            self.assertIn("output: yyy", output)
            self.assertIn("latency: 0", output)
            self.assertIn("latency: 50", output)

    def testReplaceAll(self):
        replacement = ":latency:int:50"
        output = run_command(
            f"{CMD_TOOL} --replaces {replacement} {self.storage_file_tmp}", output=True
        )
        print(output)
        with open(self.storage_file_tmp, mode="r") as opened_file:
            output = opened_file.read()
            self.assertNotIn("random_output", output)
            self.assertIn("requre.openshift_integration", output)
            self.assertNotIn("latency: xxxx", output)
            self.assertIn("output: yyy", output)
            self.assertNotIn("latency: 0", output)
            self.assertIn("latency: 50", output)


class Simplify(BaseClass):
    """
    Check simplifying of keys
    """

    keys = ["a", "b", "c", "d", "e"]
    metadata = {"latency": 0}

    def tearDown(self):
        # return it to default type
        DataMiner().data_type = DataTypes.List
        super().tearDown()

    def store_key(self):
        PersistentObjectStorage().store(
            keys=self.keys, values="x", metadata=self.metadata
        )

    def testDefault(self):
        self.store_key()
        processor = DictProcessing(PersistentObjectStorage().storage_object)
        processor.simplify()
        self.assertIn(
            "'a': {'d': {'e': [", str(PersistentObjectStorage().storage_object)
        )

    def testDict(self):
        DataMiner().data_type = DataTypes.Dict
        self.store_key()
        print(PersistentObjectStorage().storage_object)
        processor = DictProcessing(PersistentObjectStorage().storage_object)
        processor.simplify()
        self.assertIn(
            "'a': {'d': {'e': {'%s': {'metadata'" % DataMiner().key,
            str(PersistentObjectStorage().storage_object),
        )

    def testDictWithList(self):
        DataMiner().data_type = DataTypes.DictWithList
        self.store_key()
        print(PersistentObjectStorage().storage_object)
        processor = DictProcessing(PersistentObjectStorage().storage_object)
        processor.simplify()
        self.assertIn(
            "'a': {'d': {'e': {'%s': [" % DataMiner().key,
            str(PersistentObjectStorage().storage_object),
        )

    def testValue(self):
        DataMiner().data_type = DataTypes.Value
        self.store_key()
        print(PersistentObjectStorage().storage_object)
        processor = DictProcessing(PersistentObjectStorage().storage_object)
        processor.simplify()
        self.assertIn(
            "'a': {'d': {'e': {'metadata'",
            str(PersistentObjectStorage().storage_object),
        )



import builtins
import os
import sys

from requre import decorate, replace, replace_module
from requre.helpers.tempfile import TempFile
from requre.import_system import ReplaceType, upgrade_import_system, UpgradeImportSystem
from requre.storage import PersistentObjectStorage
from tests.testbase import BaseClass

SELECTOR = str(os.path.basename(__file__).rsplit(".", 1)[0])


class TestUpgradeImportSystem(BaseClass):
    def setUp(self) -> None:
        super().setUp()
        self.old_imports = builtins.__import__

    def tearDown(self) -> None:
        builtins.__import__ = self.old_imports
        if "tempfile" in sys.modules:
            del sys.modules["tempfile"]
        super().tearDown()

    def test_no_filter(self):
        old_imports = builtins.__import__
        with UpgradeImportSystem():
            import tempfile

            self.assertNotIn("decorated", tempfile.mktemp())
            self.assertIn("/tmp", tempfile.mktemp())
        assert builtins.__import__ == old_imports

    def test_import_from(self):
        """
        Test if it is able to patch also from statement
        """
        replace(
            where="^tempfile$",
            what="mktemp",
            replacement=lambda: "b",
            who_name=SELECTOR,
        )
        from tempfile import mktemp

        self.assertNotIn("/tmp", mktemp())
        self.assertIn("b", mktemp())

    def test_filter_replace(self):
        """
        Test improving of import system with import statement
        Check also debug file output if it contains proper debug data
        """
        debug_file = "__modules.log"
        HANDLE_MODULE_LIST = [
            (
                "^tempfile$",
                {"who_name": "test_import_system"},
                {"mktemp": [ReplaceType.REPLACE, lambda: "a"]},
            )
        ]
        upgrade_import_system(filters=HANDLE_MODULE_LIST, debug_file=debug_file)
        import tempfile

        self.assertNotIn("/tmp", tempfile.mktemp())
        self.assertIn("a", tempfile.mktemp())
        with open(debug_file, "r") as fd:
            output = fd.read()
            self.assertIn(SELECTOR, output)
            self.assertIn("replacing mktemp by function", output)
        os.remove(debug_file)

    def test_method_replace(self):
        """
        Test improving of import system with import statement
        Check also debug file output if it contains proper debug data
        """
        debug_file = "__modules.log"
        UpgradeImportSystem(debug_file=debug_file).replace(
            where="^tempfile$",
            what="mktemp",
            replacement=lambda: "a",
            who_name="test_import_system",
        )
        import tempfile

        self.assertNotIn("/tmp", tempfile.mktemp())
        self.assertIn("a", tempfile.mktemp())
        with open(debug_file, "r") as fd:
            output = fd.read()
            self.assertIn(SELECTOR, output)
            self.assertIn("replacing mktemp by function", output)
        os.remove(debug_file)

    def test_method_context_replace(self):
        """
        Test improving of import system with import statement
        Check also debug file output if it contains proper debug data
        """
        debug_file = "__modules.log"
        with UpgradeImportSystem(debug_file=debug_file).replace(
            where="^tempfile$",
            what="mktemp",
            replacement=lambda: "a",
            who_name="test_import_system",
        ):
            import tempfile

            self.assertNotIn("/tmp", tempfile.mktemp())
            self.assertIn("a", tempfile.mktemp())
            with open(debug_file, "r") as fd:
                output = fd.read()
                self.assertIn(SELECTOR, output)
                self.assertIn("replacing mktemp by function", output)
        os.remove(debug_file)

    def test_function_replace(self):
        """
        Test improving of import system with import statement
        Check also debug file output if it contains proper debug data
        """
        debug_file = "__modules.log"
        replace(
            where="^tempfile$",
            what="mktemp",
            replacement=lambda: "a",
            who_name="test_import_system",
            debug_file=debug_file,
        )
        import tempfile

        self.assertNotIn("/tmp", tempfile.mktemp())
        self.assertIn("a", tempfile.mktemp())
        with open(debug_file, "r") as fd:
            output = fd.read()
            self.assertIn(SELECTOR, output)
            self.assertIn("replacing mktemp by function", output)
        os.remove(debug_file)

    def test_function_context_replace(self):
        """
        Test improving of import system with import statement
        Check also debug file output if it contains proper debug data
        """
        debug_file = "__modules.log"
        with replace(
            where="^tempfile$",
            what="mktemp",
            replacement=lambda: "a",
            who_name="test_import_system",
            debug_file=debug_file,
        ):
            import tempfile

            self.assertNotIn("/tmp", tempfile.mktemp())
            self.assertIn("a", tempfile.mktemp())
            with open(debug_file, "r") as fd:
                output = fd.read()
                self.assertIn(SELECTOR, output)
                self.assertIn("replacing mktemp by function", output)
            os.remove(debug_file)

    def test_filter_decorate(self):
        """
        Test patching by decorator_all_keys, not replacing whole function
        """
        HANDLE_MODULE_LIST = [
            (
                "^tempfile$",
                {"who_name": SELECTOR},
                {
                    "mktemp": [
                        ReplaceType.DECORATOR,
                        lambda x: lambda: f"decorated {x()}",
                    ]
                },
            )
        ]
        upgrade_import_system(filters=HANDLE_MODULE_LIST)
        import tempfile

        self.assertIn("decorated", tempfile.mktemp())
        self.assertIn("/tmp", tempfile.mktemp())

    def test_method_decorate(self):
        """
        Test patching by decorator_all_keys, not replacing whole function
        """
        UpgradeImportSystem().decorate(
            where="^tempfile$",
            what="mktemp",
            decorator=lambda x: lambda: f"decorated {x()}",
            who_name=SELECTOR,
        )

        import tempfile

        self.assertIn("decorated", tempfile.mktemp())
        self.assertIn("/tmp", tempfile.mktemp())

    def test_method_context_decorate(self):
        """
        Test patching by decorator_all_keys, not replacing whole function
        """
        with UpgradeImportSystem().decorate(
            where="^tempfile$",
            what="mktemp",
            decorator=lambda x: lambda: f"decorated {x()}",
            who_name=SELECTOR,
        ):
            import tempfile

            self.assertIn("decorated", tempfile.mktemp())
            self.assertIn("/tmp", tempfile.mktemp())

    def test_function_decorate(self):
        """
        Test patching by decorator_all_keys, not replacing whole function
        """
        decorate(
            where="^tempfile$",
            what="mktemp",
            decorator=lambda x: lambda: f"decorated {x()}",
            who_name=SELECTOR,
        )

        import tempfile

        self.assertIn("decorated", tempfile.mktemp())
        self.assertIn("/tmp", tempfile.mktemp())

    def test_function_context_decorate(self):
        """
        Test patching by decorator_all_keys, not replacing whole function
        """
        with decorate(
            where="^tempfile$",
            what="mktemp",
            decorator=lambda x: lambda: f"decorated {x()}",
            who_name=SELECTOR,
        ):
            import tempfile

            self.assertIn("decorated", tempfile.mktemp())
            self.assertIn("/tmp", tempfile.mktemp())

    def test_filter_replace_module(self):
        """
        Test if it is able to replace whole module by own implemetation
        Test also own implementation of static tempfile module via class
        """

        HANDLE_MODULE_LIST = [
            (
                "^tempfile$",
                {"who_name": SELECTOR},
                {"": [ReplaceType.REPLACE_MODULE, TempFile]},
            )
        ]
        upgrade_import_system(filters=HANDLE_MODULE_LIST)
        import tempfile

        tmpfile = tempfile.mktemp()
        tmpdir = tempfile.mkdtemp()
        self.assertIn(
            f"/tmp/{os.path.basename(PersistentObjectStorage().storage_file)}/static_tmp",
            tmpfile,
        )
        self.assertIn(
            f"/tmp/{os.path.basename(PersistentObjectStorage().storage_file)}/static_tmp",
            tmpdir,
        )
        self.assertFalse(os.path.exists(tmpfile))
        self.assertTrue(os.path.exists(tmpdir))
        self.assertTrue(os.path.isdir(tmpdir))
        os.removedirs(tmpdir)

    def test_method_replace_module(self):
        """
        Test if it is able to replace whole module by own implemetation
        Test also own implementation of static tempfile module via class
        """

        UpgradeImportSystem().replace_module(
            where="^tempfile$", replacement=TempFile, who_name=SELECTOR
        )

        import tempfile

        tmpfile = tempfile.mktemp()
        tmpdir = tempfile.mkdtemp()
        self.assertIn(
            f"/tmp/{os.path.basename(PersistentObjectStorage().storage_file)}/static_tmp",
            tmpfile,
        )
        self.assertIn(
            f"/tmp/{os.path.basename(PersistentObjectStorage().storage_file)}/static_tmp",
            tmpdir,
        )
        self.assertFalse(os.path.exists(tmpfile))
        self.assertTrue(os.path.exists(tmpdir))
        self.assertTrue(os.path.isdir(tmpdir))
        os.removedirs(tmpdir)

    def test_method_context_replace_module(self):
        """
        Test if it is able to replace whole module by own implemetation
        Test also own implementation of static tempfile module via class
        """

        with UpgradeImportSystem().replace_module(
            where="^tempfile$", replacement=TempFile, who_name=SELECTOR
        ):
            import tempfile

            tmpfile = tempfile.mktemp()
            tmpdir = tempfile.mkdtemp()
            self.assertIn(
                f"/tmp/{os.path.basename(PersistentObjectStorage().storage_file)}/static_tmp",
                tmpfile,
            )
            self.assertIn(
                f"/tmp/{os.path.basename(PersistentObjectStorage().storage_file)}/static_tmp",
                tmpdir,
            )
            self.assertFalse(os.path.exists(tmpfile))
            self.assertTrue(os.path.exists(tmpdir))
            self.assertTrue(os.path.isdir(tmpdir))
            os.removedirs(tmpdir)

    def test_function_replace_module(self):
        """
        Test if it is able to replace whole module by own implemetation
        Test also own implementation of static tempfile module via class
        """

        replace_module(where="^tempfile$", replacement=TempFile, who_name=SELECTOR)
        import tempfile

        tmpfile = tempfile.mktemp()
        tmpdir = tempfile.mkdtemp()
        self.assertIn(
            f"/tmp/{os.path.basename(PersistentObjectStorage().storage_file)}/static_tmp",
            tmpfile,
        )
        self.assertIn(
            f"/tmp/{os.path.basename(PersistentObjectStorage().storage_file)}/static_tmp",
            tmpdir,
        )
        self.assertFalse(os.path.exists(tmpfile))
        self.assertTrue(os.path.exists(tmpdir))
        self.assertTrue(os.path.isdir(tmpdir))
        os.removedirs(tmpdir)

    def test_function_context_replace_module(self):
        """
        Test if it is able to replace whole module by own implemetation
        Test also own implementation of static tempfile module via class
        """

        with replace_module(
            where="^tempfile$", replacement=TempFile, who_name=SELECTOR
        ):
            import tempfile

            tmpfile = tempfile.mktemp()
            tmpdir = tempfile.mkdtemp()
            self.assertIn(
                f"/tmp/{os.path.basename(PersistentObjectStorage().storage_file)}/static_tmp",
                tmpfile,
            )
            self.assertIn(
                f"/tmp/{os.path.basename(PersistentObjectStorage().storage_file)}/static_tmp",
                tmpdir,
            )
            self.assertFalse(os.path.exists(tmpfile))
            self.assertTrue(os.path.exists(tmpdir))
            self.assertTrue(os.path.isdir(tmpdir))
            os.removedirs(tmpdir)
